import requests


def curl(jenkins_user, jenkins_token, jenkins_url, post_json_data=None, raw=False):
    res = requests.post(jenkins_url, auth=(jenkins_user, jenkins_token), json=post_json_data)
    if raw:
        return res.text
    else:
        return res.json()


def get_session(jenkins_user, jenkins_token):
    session = requests.session()
    session.auth = (jenkins_user, jenkins_token)
    return session



import traceback

from ckan_cloud_operator.datastore_permissions import DATASTORE_PERMISSIONS_SQL_TEMPLATE
from ckan_cloud_operator.providers.db import manager as db_manager
from ckan_cloud_operator.providers.ckan.db import migration as ckan_db_migration_manager
from ckan_cloud_operator.drivers.postgres import driver as postgres_driver


class DeisCkanInstanceDb(object):

    def __init__(self, instance, db_type):
        self.instance = instance
        assert db_type in ['db', 'datastore']
        self.db_type = db_type
        self.db_spec = getattr(self.instance.spec, db_type)
        self.db_prefix = self.db_spec.get('dbPrefix')

    def set_datastore_readonly_permissions(self):
        assert self.db_type == 'datastore'
        db_name = self.db_spec['name']
        ro_user = self.instance.annotations.get_secret('datastoreReadonlyUser')
        print(f'setting datastore permissions: {db_name} ({ro_user})')
        with postgres_driver.connect(db_manager.get_external_admin_connection_string(db_name=db_name, db_prefix=self.db_prefix)) as conn:
            with conn.cursor() as cur:
                for line in [
                    f"GRANT CONNECT ON DATABASE \"{db_name}\" TO \"{ro_user}\";",
                    f"GRANT USAGE ON SCHEMA public TO \"{ro_user}\";",
                    f"GRANT SELECT ON ALL TABLES IN SCHEMA public TO \"{ro_user}\";",
                    f"ALTER DEFAULT PRIVILEGES FOR USER \"{db_name}\" IN SCHEMA public GRANT SELECT ON TABLES TO \"{ro_user}\";",
                ]:
                    cur.execute(line)
                # site_user = self.instance.spec.db['name']
                site_user = db_name
                datastore_permissions = DATASTORE_PERMISSIONS_SQL_TEMPLATE.replace('{{SITE_USER}}', site_user).replace('{{DS_RO_USER}}', ro_user)
                try:
                    cur.execute(datastore_permissions)
                except Exception:
                    traceback.print_exc()
                    print('Failed to set datastore sql template, continuing anyway')

    def update(self):
        db_migration_name = self.db_spec.get('fromDbMigration')
        if not db_migration_name:
            db_migration_name = f'new-{self.db_type}-{self.instance.id}'
            db_migration_spec = {
                'type': f'new-{self.db_type}',
                f'{self.db_type}-name': self.db_spec['name'],
                **({'db-prefix': self.db_spec['dbPrefix']} if self.db_spec.get('dbPrefix') else {}),
            }
            migration = ckan_db_migration_manager.create(
                db_migration_name, db_migration_spec, exists_ok=True
            )
            for event in ckan_db_migration_manager.update(migration):
                ckan_db_migration_manager.print_event_exit_on_complete(event, '')
        database_password, datastore_password, datastore_readonly_password = ckan_db_migration_manager.get_dbs_passwords(db_migration_name)
        datastore_readonly_user = ckan_db_migration_manager.get_datastore_raedonly_user_name(db_migration_name)
        if self.db_type == 'db':
            if database_password:
                self.instance.annotations.set_secrets({
                    'databasePassword': database_password
                })
        elif self.db_type == 'datastore':
            if datastore_password and datastore_readonly_password:
                self.instance.annotations.set_secrets({
                    'datastorePassword': datastore_password,
                    'datastoreReadonlyUser': datastore_readonly_user,
                    'datatastoreReadonlyPassword': datastore_readonly_password,
                })

    def get(self):
        return {'ready': db_manager.check_db_exists(self.db_spec['name'], db_prefix=self.db_prefix),
                'db-name': self.db_spec['name'],
                'db-prefix': self.db_prefix}



PROVIDER_ID = 'solrcloud'


LOG4J_PROPERTIES = """
# Logging level
solr.log=logs
log4j.rootLogger=WARN, file, CONSOLE

log4j.appender.CONSOLE=org.apache.log4j.ConsoleAppender

log4j.appender.CONSOLE.layout=org.apache.log4j.EnhancedPatternLayout
log4j.appender.CONSOLE.layout.ConversionPattern=%-4r %-5p (%t) [%X{collection} %X{shard} %X{replica} %X{core}] %c{1.} %m%n

#- size rotation with log cleanup.
log4j.appender.file=org.apache.log4j.RollingFileAppender
log4j.appender.file.MaxFileSize=4MB
log4j.appender.file.MaxBackupIndex=9

#- File to log to and log format
log4j.appender.file.File=${solr.log}/solr.log
log4j.appender.file.layout=org.apache.log4j.EnhancedPatternLayout
log4j.appender.file.layout.ConversionPattern=%d{yyyy-MM-dd HH:mm:ss.SSS} %-5p (%t) [%X{collection} %X{shard} %X{replica} %X{core}] %c{1.} %m%n

log4j.logger.org.apache.zookeeper=WARN
log4j.logger.org.apache.hadoop=WARN

# set to INFO to enable infostream log messages
log4j.logger.org.apache.solr.update.LoggingInfoStream=OFF
"""


SOLR_CONFIG_XML = """<?xml version="1.0" encoding="UTF-8" ?>
<solr>
  <solrcloud>
    <str name="host">${host:}</str>
    <int name="hostPort">${jetty.port:8983}</int>
    <str name="hostContext">${hostContext:solr}</str>
    <bool name="genericCoreNodeNames">${genericCoreNodeNames:true}</bool>
    <int name="zkClientTimeout">${zkClientTimeout:30000}</int>
    <int name="distribUpdateSoTimeout">${distribUpdateSoTimeout:600000}</int>
    <int name="distribUpdateConnTimeout">${distribUpdateConnTimeout:60000}</int>
  </solrcloud>
  <shardHandlerFactory name="shardHandlerFactory"
    class="HttpShardHandlerFactory">
    <int name="socketTimeout">${socketTimeout:600000}</int>
    <int name="connTimeout">${connTimeout:60000}</int>
  </shardHandlerFactory>
</solr>"""



import click
import json

from ckan_cloud_operator import logs
from .deployment import cli as deployment_cli
from . import manager


@click.group()
def apps():
    """Manage Generic Application Instances"""
    pass


apps.add_command(deployment_cli.deployment)


@apps.command()
def initialize():
    manager.initialize()
    logs.exit_great_success()


@apps.command()
@click.argument('VALUES_FILE')
@click.option('--instance-id')
@click.option('--instance-name')
@click.option('--exists-ok', is_flag=True)
@click.option('--dry-run', is_flag=True)
@click.option('--update', 'update_', is_flag=True)
@click.option('--wait-ready', is_flag=True)
@click.option('--skip-deployment', is_flag=True)
@click.option('--skip-route', is_flag=True)
@click.option('--force', is_flag=True)
def create(values_file, instance_id, instance_name, exists_ok, dry_run, update_, wait_ready,
           skip_deployment, skip_route, force):
    manager.create(instance_id=instance_id, instance_name=instance_name,
                   values_filename=values_file, exists_ok=exists_ok, dry_run=dry_run, update_=update_,
                   wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route, force=force)
    logs.exit_great_success()


@apps.command()
@click.argument('INSTANCE_ID_OR_NAME')
@click.argument('OVERRIDE_SPEC_JSON', required=False)
@click.option('--persist-overrides', is_flag=True)
@click.option('--wait-ready', is_flag=True)
@click.option('--skip-deployment', is_flag=True)
@click.option('--skip-route', is_flag=True)
@click.option('--force', is_flag=True)
def update(instance_id_or_name, override_spec_json, persist_overrides, wait_ready, skip_deployment, skip_route, force):
    """Update an instance to the latest resource spec, optionally applying the given json override to the resource spec

    Examples:

    ckan-cloud-operator apps update <INSTANCE_ID_OR_NAME> '{"siteUrl": "http://localhost:5000"}' --wait-ready

    ckan-cloud-operator apps update <INSTANCE_ID_OR_NAME> '{"replicas": 3}' --persist-overrides
    """
    override_spec = json.loads(override_spec_json) if override_spec_json else None
    manager.update(instance_id_or_name, override_spec=override_spec, persist_overrides=persist_overrides,
                   wait_ready=wait_ready, skip_deployment=skip_deployment, skip_route=skip_route,
                   force=force)
    logs.exit_great_success()


@apps.command()
@click.argument('INSTANCE_ID_OR_NAME')
@click.argument('ATTR', required=False)
@click.option('--with-spec', is_flag=True)
def get(instance_id_or_name, attr, with_spec):
    """Get detailed information about an instance, optionally returning only a single get attribute

    Example: ckan-cloud-operator apps get <INSTANCE_ID_OR_NAME> deployment
    """
    if attr == 'spec':
        with_spec = True
    logs.print_yaml_dump(manager.get(instance_id_or_name, attr, with_spec=with_spec), exit_success=True)


@apps.command()
@click.argument('INSTANCE_ID_OR_NAME')
def edit(instance_id_or_name):
    manager.edit(instance_id_or_name)


@apps.command('list')
@click.option('-f', '--full', is_flag=True)
@click.option('-q', '--quick', is_flag=True)
@click.option('--name')
def list_instances(full, quick, name):
    for instance in manager.list_instances(full=full, quick=quick, name=name):
        logs.print_yaml_dump([instance])
    logs.exit_great_success(quiet=True)


@apps.command()
@click.argument('INSTANCE_ID')
@click.argument('INSTANCE_NAME')
def set_name(instance_id, instance_name):
    logs.info(f'{instance_name} --> {instance_id}')
    manager.set_name(instance_id, instance_name)
    logs.exit_great_success()


@apps.command()
@click.argument('INSTANCE_NAME')
def delete_name(instance_name):
    manager.delete_name(instance_name=instance_name)
    logs.exit_great_success()


@apps.command()
@click.argument('INSTANCE_ID_OR_NAME', nargs=-1)
@click.option('--no-dry-run', is_flag=True)
def delete(instance_id_or_name, no_dry_run):
    generator = manager.delete_instances(instance_ids_or_names=instance_id_or_name, dry_run=not no_dry_run)
    while True:
        try:
            logs.info('Deleting instance', **next(generator))
        except StopIteration:
            break
        logs.info(**next(generator))
    logs.exit_great_success(quiet=True)



from ckan_cloud_operator.providers import manager as providers_manager
from .constants import PROVIDER_SUBMODULE
from .adminer.constants import PROVIDER_ID as adminer_provider_id


def initialize():
    get_provider(default=adminer_provider_id).initialize()


def start():
    """Start a web-UI for db management"""
    get_provider().start()


def get_provider(default=None):
    return providers_manager.get_provider(PROVIDER_SUBMODULE, default=default)



class DeisCkanInstanceStorage(object):

    def __init__(self, instance):
        self.instance = instance
        self.storage_spec = self.instance.spec.storage
        self.storage_path = self.storage_spec['path']

    def update(self):
        assert self.get()['ready']

    def delete(self):
        print('WARNING! Storage was not deleted')

    def get(self):
        storage_path = self.storage_spec['path']
        from ckan_cloud_operator.providers.ckan.storage import manager as ckan_storage_manager
        return ckan_storage_manager.get_storage_path_status(storage_path)



import click

from ckan_cloud_operator import logs

from . import manager


@click.group()
def helm():
    """Manage App instances deployed using Helm"""
    pass


@helm.command()
def initialize():
    logs.exit_great_success()


@helm.command()
@click.option('--tiller-namespace-name')
@click.option('--chart-repo')
@click.option('--chart-version')
@click.option('--chart-release-name')
@click.option('--values-json')
@click.option('--values-filename')
@click.option('--instance-id')
@click.option('--instance-name')
@click.option('--exists-ok', is_flag=True)
@click.option('--dry-run', is_flag=True)
@click.option('--update', 'update_', is_flag=True)
@click.option('--wait-ready', is_flag=True)
@click.option('--skip-deployment', is_flag=True)
@click.option('--skip-route', is_flag=True)
@click.option('--force', is_flag=True)
@click.option('--app-type')
def create(**kwargs):
    instance_id = manager.create(**kwargs)
    logs.info(f'\n\nCreated instance ID: \n\n{instance_id}\n\n')
    logs.exit_great_success()



import subprocess
import json

from ckan_cloud_operator import logs


def get_zone_id(auth_email, auth_key, zone_name):
    data = curl(auth_email, auth_key, 'zones')
    zones = [zone['id'] for zone in data['result'] if zone['name'] == zone_name]
    return zones[0] if len(zones) > 0 else None


def get_zone_rate_limits(auth_email, auth_key, zone_name):
    zone_id = get_zone_id(auth_email, auth_key, zone_name)
    return curl(auth_email, auth_key, f'zones/{zone_id}/rate_limits?page=1&per_page=1000')


def get_record_id(auth_email, auth_key, zone_id, record_name):
    data = curl(auth_email, auth_key, f'zones/{zone_id}/dns_records?name={record_name}')
    records = [record['id'] for record in data['result'] if record['name'] == record_name]
    return records[0] if len(records) > 0 else None


def is_ip(target_ip):
    return all([c in '0123456789.' for c in target_ip.strip()])


def update_a_record(auth_email, auth_key, zone_name, record_name, target_ip):
    zone_id = get_zone_id(auth_email, auth_key, zone_name)
    assert zone_id is not None, f'Invalid zone name: {zone_name}'
    record_id = get_record_id(auth_email, auth_key, zone_id, record_name)

    cf_record = {'type': 'A' if is_ip(target_ip) else 'CNAME',
                 'name': record_name, 'content': target_ip, 'ttl': 120, 'proxied': False}

    if record_id:
        print(f'Updating existing record {record_name}')
        data = curl(auth_email, auth_key, f'zones/{zone_id}/dns_records/{record_id}', cf_record, 'PUT')
    else:
        print(f'Creating new record: {record_name}')
        data = curl(auth_email, auth_key, f'zones/{zone_id}/dns_records', cf_record, 'POST')
    assert data.get('success')


def curl(auth_email, auth_key, urlpart, data=None, method='GET'):
    logs.info(f'Running Cloudflare curl: {urlpart} {data} {method}')
    logs.debug(f'{auth_email} / {auth_key}')
    cmd = ['curl', '-s', '-X', method, f'https://api.cloudflare.com/client/v4/{urlpart}']
    cmd += [
        '-H', f'X-Auth-Email: {auth_email}',
        '-H', f'X-Auth-Key: {auth_key}',
        '-H', 'Content-Type: application/json'
    ]
    if data:
        cmd += ['--data', json.dumps(data)]
    logs.debug(*cmd)
    output = subprocess.check_output(cmd)
    try:
        return json.loads(output)
    except Exception:
        logs.critical(f'Got invalid data from cloudflare curl: {output}')
        raise



from unittest.mock import patch

from ckan_cloud_operator.cli import deis_instance
from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance, DeisCkanInstanceCKAN
from tests.bases import BaseCliTestCase


class DeisInstanceCliTestCase(BaseCliTestCase):
    @patch.object(DeisCkanInstance, 'get')
    def test_get(self, get):
        self.runner.invoke(deis_instance, ['get', 'montreal', 'deployment'])
        get.assert_called_once_with('deployment')

    @patch.object(DeisCkanInstance, 'create')
    def test_create_from_gitlab(self, create):
        self.runner.invoke(deis_instance, ['create', 'from-gitlab', 'viderum/cloud-demo2', 'ckan_27_default', 'montreal', '--db-prefix=prod2'])
        create.assert_called_once_with(
            'from-gitlab',
            'viderum/cloud-demo2',
            'ckan_27_default',
            'montreal',
            no_db_proxy=False,
            storage_path=None,
            from_db_backups=None,
            solr_collection=None,
            rerun=False,
            force=False,
            recreate_dbs=False,
            db_prefix='prod2',
            use_private_gitlab_repo=False
        )

    @patch.object(DeisCkanInstance, 'update')
    @patch('ckan_cloud_operator.deis_ckan.instance.ckan_manager')
    @patch('ckan_cloud_operator.deis_ckan.instance.subprocess.call')
    def test_edit(self, subprocess_call, ckan_manager, update):
        ckan_manager.instance_kind.return_value = 'ckancloudckaninstance'
        self.runner.invoke(deis_instance, ['edit', 'montreal', 'vim'])
        subprocess_call.assert_called_with('EDITOR=vim kubectl -n ckan-cloud edit ckancloudckaninstance/montreal', shell=True)
        self.assertEqual(update.call_count, 1)

    @patch.object(DeisCkanInstance, 'delete')
    def test_delete(self, delete):
        self.runner.invoke(deis_instance, ['delete', 'montreal'])
        self.assertEqual(delete.call_count, 1)

    @patch.object(DeisCkanInstance, 'delete')
    def test_delete_multiple(self, delete):
        self.runner.invoke(deis_instance, ['delete', 'montreal', 'another_instance', 'new_instance'])
        self.assertEqual(delete.call_count, 3)

    @patch.object(DeisCkanInstance, 'list')
    def test_list(self, deis_list):
        self.runner.invoke(deis_instance, ['list', '-f', '-q'])
        deis_list.assert_called_once_with(True, True)



import json
import unittest
from unittest.mock import patch, call

from ckan_cloud_operator.providers.solr import manager


class ZooKeeperTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.providers.solr.manager.kubectl.get')
    @patch('ckan_cloud_operator.providers.solr.manager.kubectl.check_output')
    def test_put_configs(self, kubectl_check_output, kubectl_get):
        kubectl_get.return_value = {
            'items': [
                {
                    'metadata': {
                        'name': 'zk-default-pod'
                    }
                }
            ]
        }
        manager.zk_put_configs('tests/test_data/schema')
        self.assertEqual(kubectl_check_output.call_count, 3)
        expected_calls = [
            call('exec zk-default-pod zkCli.sh create /configs null'),
            call('cp tests/test_data/schema/schema.xml zk-default-pod:/tmp/zk_input'),
            call('exec zk-default-pod bash -- -c \'zkCli.sh create /configs/schema.xml "$(cat /tmp/zk_input)"\'')
        ]
        self.assertEqual(kubectl_check_output.call_args_list, expected_calls)



#### standard provider code ####

from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)
def _config_get_volume_spec(volume_name, is_secret=False, suffix=None): return providers_manager.config_get_volume_spec(PROVIDER_SUBMODULE, PROVIDER_ID, volume_name, is_secret, suffix)

################################
# custom provider code starts here
#

import os
import subprocess

from distutils.util import strtobool

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.config import manager as config_manager


def initialize(db_prefix=None):
    _apply_service(db_prefix)
    _apply_deployment(db_prefix)
    if not db_prefix:
        _set_provider()
    else:
        _config_interactive_set({
            'port-forwarded-port': '5433'
        }, suffix=db_prefix)
        db_prefixes = _config_get('all-db-prefixes')
        db_prefixes = db_prefixes.split(',') if db_prefixes else []
        if db_prefix not in db_prefixes:
            db_prefixes.append(db_prefix)
            _config_set('all-db-prefixes', ','.join(db_prefixes))


def get_all_db_prefixes():
    prefixes = _config_get('all-db-prefixes')
    if prefixes:
        return prefixes.split(',')
    else:
        return []


def update(**kwargs):
    pass


def get_internal_proxy_host_port(db_prefix=None):
    namespace = cluster_manager.get_operator_namespace_name()
    service_name = _get_resource_name(suffix=db_prefix or '')
    return f'{service_name}.{namespace}', 5432


def get_external_proxy_host_port(db_prefix=None):
    if strtobool(os.environ.get('CKAN_CLOUD_OPERATOR_USE_PROXY', 'y')):
        host, port = get_external_proxy_forwarded_host_port(db_prefix)
    else:
        host, port = None, None
    return host, port


def get_external_proxy_forwarded_host_port(db_prefix=None):
    return '127.0.0.1', get_port_forwarded_port(db_prefix)


def get_port_forwarded_port(db_prefix=None):
    if db_prefix:
        return _config_get(key='port-forwarded-port', suffix=db_prefix or '')
    else:
        return 5432


def start_port_forward(db_prefix=None):
    """Starts a local proxy to the cloud SQL instance"""
    print("\nKeep this running in the background\n")
    namespace = cluster_manager.get_operator_namespace_name()
    deployment_name = _get_resource_name(suffix=db_prefix or '')
    port = get_port_forwarded_port(db_prefix)
    subprocess.check_call(f'kubectl -n {namespace} port-forward deployment/{deployment_name} {port}:5432',
                          shell=True)


def _apply_deployment(db_prefix=None):
    rds_host = config_manager.get(secret_name='ckan-cloud-provider-db-rds-credentials')['rds-host']
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(suffix=db_prefix),
        _get_resource_labels(for_deployment=True, suffix=db_prefix or ''),
        {
            'replicas': 1,
            'revisionHistoryLimit': 10,
            'strategy': {'type': 'RollingUpdate', },
            'selector': {'matchLabels': _get_resource_labels(for_deployment=True, suffix=db_prefix or '')},
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix=db_prefix or ''),
                    'annotations': _get_resource_annotations(suffix=db_prefix or '')
                },
                'spec': {
                    'containers': [
                        {
                            'name': 'proxy',
                            # https://github.com/OriHoch/docker-tcp-proxy
                            'image': 'viderum/docker-tcp-proxy:latest',
                            'env': [
                                {
                                    'name': 'LISTEN',
                                    'value': ':5432'
                                },
                                {
                                    'name': 'TALK',
                                    'value': f'{rds_host}:5432',
                                }
                            ],
                            'ports': [{'containerPort': 5432}],
                            'resources': {
                                'limits': {
                                    'memory': '0.5Gi',
                                },
                                'requests': {
                                    'cpu': '0.1',
                                    'memory': '0.2Gi',
                                }
                            }
                        }
                    ]
                }
            }
        }
    ))


def _apply_service(db_prefix=None):
    deployment_app = _get_resource_labels(for_deployment=True, suffix=db_prefix or '')['app']
    kubectl.apply(kubectl.get_service(
        _get_resource_name(suffix=db_prefix or ''),
        _get_resource_labels(suffix=db_prefix or ''),
        [5432],
        {'app': deployment_app}
    ))



import click
import binascii
import os
import yaml

from ckan_cloud_operator import logs

from . import driver


@click.group()
def postgres():
    """Manage PostgreSQL databases unrelated to the operator or cluster"""
    pass


@postgres.command()
@click.option('--db-name')
@click.option('--db-password')
@click.option('--admin-connection-string')
def create_base_db(db_name, db_password, admin_connection_string):
    with _get_admin_connection(admin_connection_string) as admin_conn:
        db_name = db_name or _generate_password(8)
        db_password = db_password or _generate_password(12)
        logs.info(f'creating base db: {db_name} / {db_password}')
        driver.create_base_db(admin_conn, db_name, db_password)
        logs.exit_great_success()


@postgres.command()
@click.option('--db-name')
@click.option('--admin-connection-string')
def db_role_info(db_name, admin_connection_string):
    assert db_name
    with _get_admin_connection(admin_connection_string) as admin_conn:
        print(yaml.dump(driver.get_db_role_info(admin_conn, db_name), default_flow_style=False))


@postgres.command()
@click.option('--admin-connection-string')
@click.option('--full', is_flag=True)
@click.option('--validate', is_flag=True)
def db_names(admin_connection_string, full, validate):
    with _get_admin_connection(admin_connection_string) as admin_conn:
        print(yaml.dump(list(driver.list_db_names(admin_conn, full, validate)), default_flow_style=False))


@postgres.command()
@click.option('--admin-connection-string')
@click.option('--full', is_flag=True)
@click.option('--validate', is_flag=True)
def roles(admin_connection_string, full, validate):
    with _get_admin_connection(admin_connection_string) as admin_conn:
        print(yaml.dump(list(driver.list_roles(admin_conn, full, validate)), default_flow_style=False))


def _get_admin_connection_string(connection_string):
    if connection_string: return connection_string
    from ckan_cloud_operator.providers.db.manager import get_external_admin_connection_string
    return get_external_admin_connection_string()


def _get_admin_connection(connection_string):
    return driver.connect(_get_admin_connection_string(connection_string))


def _generate_password(size):
    return binascii.hexlify(os.urandom(size)).decode()



import unittest
from unittest.mock import patch, Mock, MagicMock, call

from ckan_cloud_operator.deis_ckan.ckan import DeisCkanInstanceCKAN
from ckan_cloud_operator.deis_ckan.instance import DeisCkanInstance
from ckan_cloud_operator.deis_ckan.db import DeisCkanInstanceDb, postgres_driver, db_manager, DATASTORE_PERMISSIONS_SQL_TEMPLATE


class CkanTestCase(unittest.TestCase):
    def setUp(self):
        self.ckan = DeisCkanInstanceCKAN(DeisCkanInstance('montreal'))

    def test_run_exec_command(self):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.run('exec', 'ls', '-lrt')
        self.ckan.instance.kubectl.assert_called_once_with('exec montreal-asd-qwe ls -lrt', check_output=False)

    def test_run_paster_command(self):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.run('paster', 'initdb')
        self.ckan.instance.kubectl.assert_called_once_with('exec montreal-asd-qwe -it -- paster --plugin=ckan initdb -c /srv/app/production.ini ', check_output=False)

    def test_run_logs_command(self):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.run('logs')
        self.ckan.instance.kubectl.assert_called_once_with('logs montreal-asd-qwe ')

    @patch('ckan_cloud_operator.deis_ckan.ckan.subprocess.check_call')
    def test_run_port_forward_command(self, check_call):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.run('port-forward', '5001')
        self.ckan.instance.kubectl.assert_called_once_with('port-forward montreal-asd-qwe 5001')
        check_call.assert_called_once_with(['kubectl', '-n', 'montreal', 'port-forward', 'deployment/montreal', '5001'])

    def test_run_with_unsupported_command(self):
        with self.assertRaises(AssertionError):
            self.ckan.run('delete', 'everything')

    def test_init_creation_command(self):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.instance._spec = MagicMock()
        self.ckan.instance.spec.spec = {
            'ckan': {
                'init': [
                    ['paster', 'initdb'],
                    ['paster', 'test', '-v']
                ]
            }
        }
        self.ckan._create()
        self.assertEqual(self.ckan.instance.kubectl.call_count, 2)
        self.ckan.instance.kubectl.assert_has_calls([
            call('exec montreal-asd-qwe -it -- paster --plugin=ckan initdb -c /srv/app/production.ini ', check_output=False),
            call('exec montreal-asd-qwe -it -- paster --plugin=ckan test -c /srv/app/production.ini -v', check_output=False)
        ])

    def test_init_creation_with_non_paster_command(self):
        self.ckan._get_ckan_pod_name = lambda: 'montreal-asd-qwe'
        self.ckan.instance.kubectl = MagicMock()
        self.ckan.instance._spec = MagicMock()
        self.ckan.instance.spec.spec = {
            'ckan': {
                'init': [
                    ['paster', 'initdb'],
                    ['ls', '-lrt']
                ]
            }
        }
        with self.assertRaisesRegex(ValueError, "Invalid ckan init cmd: ['ls', '-lrt']"):
            self.ckan._create()
        self.ckan.instance.kubectl.assert_called_once_with('exec montreal-asd-qwe -it -- paster --plugin=ckan initdb -c /srv/app/production.ini ', check_output=False)


class DbTestCase(unittest.TestCase):
    def setUp(self):
        self.instance = DeisCkanInstance('montreal')
        self.instance._spec = Mock(datastore={'name': 'test'})
        self.db = DeisCkanInstanceDb(self.instance, 'db')
        self.datastore = DeisCkanInstanceDb(self.instance, 'datastore')

    def test_set_datastore_readonly_permissions(self):
        postgres_driver.connect = MagicMock()
        db_manager.get_external_admin_connection_string = MagicMock()
        conn = MagicMock()
        cursor = MagicMock()
        conn.cursor.return_value.__enter__.return_value = cursor
        postgres_driver.connect.return_value.__enter__.return_value = conn
        self.instance.annotations.get_secret = lambda x: 'postgres_ro'

        self.datastore.set_datastore_readonly_permissions()

        self.assertEqual(cursor.execute.call_count, 5)
        cursor.execute.assert_has_calls([
            call("GRANT CONNECT ON DATABASE \"test\" TO \"postgres_ro\";"),
            call("GRANT USAGE ON SCHEMA public TO \"postgres_ro\";"),
            call("GRANT SELECT ON ALL TABLES IN SCHEMA public TO \"postgres_ro\";"),
            call("ALTER DEFAULT PRIVILEGES FOR USER \"test\" IN SCHEMA public GRANT SELECT ON TABLES TO \"postgres_ro\";"),
            call(DATASTORE_PERMISSIONS_SQL_TEMPLATE.replace('{{SITE_USER}}', 'test').replace('{{DS_RO_USER}}', 'postgres_ro'))
        ])

    def test_set_datastore_readonly_permissions_for_db(self):
        with self.assertRaises(AssertionError):
            self.db.set_datastore_readonly_permissions()



import subprocess
import yaml

from ckan_cloud_operator import logs

from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.infra import CkanInfra


def initialize(interactive=False):
    config_manager.interactive_set(
        {
            'default-storage-bucket': 'ckan',
        },
        secret_name='ckan-storage-config',
        interactive=interactive
    )
    logs.warning('Minio bucket policy was not applied!')


def get_deis_minio_bucket_policy():
    from ckan_cloud_operator.providers.ckan import manager as ckan_manager
    path_to_old_cluster_kubeconfig = ckan_manager.get_path_to_old_cluster_kubeconfig()
    return subprocess.check_output(f'KUBECONFIG={path_to_old_cluster_kubeconfig} '
                                   f'kubectl -n deis exec -it deis-minio-6ddd8f5d85-wphhb '
                                   f'-- cat /export/.minio.sys/buckets/ckan/policy.json',
                                   shell=True).decode()


def get_default_storage_bucket():
    return config_get('default-storage-bucket')


def config_get(key):
    return config_manager.get(key, secret_name='ckan-storage-config')


def get_storage_path_status(storage_path, storage_bucket=None):
    return {'ready': True, 'path': storage_path, 'bucket': storage_bucket or get_default_storage_bucket()}
    # if not storage_bucket: storage_bucket = get_default_storage_bucket()
    # Some of the srotages are not named after instance, but {instance}-prod/stage etc..
    # prod_paths = [storage_path, f'{storage_path}-prod', f'{storage_path}-production']
    # staging_paths = [storage_path, f'{storage_path}-stage', f'{storage_path}-staging']
    # returncode, output = None, None
    # for p_path, s_path in zip(prod_paths, staging_paths):
    #     storage_path_ = s_path if 'stage' in 'storage_path' else p_path
        # returncode, output = cluster_manager.get_provider().getstatusoutput(
        #     f'ls gs://{storage_bucket}{storage_path_}', gsutil=True
        # )
        # if returncode == 0:
        #     break
    # if returncode == 0:
    #     return {'ready': True, 'path': storage_path, 'bucket': storage_bucket,
    #             'root_paths': [l.strip() for l in output.split('\n')]}
    # else:
    #     return {'ready': False, 'path': storage_path, 'bucket': storage_bucket,
    #             'error': output}


def update(storage_path, storage_bucket=None):
    data = get_storage_path_status(storage_path, storage_bucket)
    ready, storage_path, storage_bucket = data['ready'], data['path'], data['bucket']
    if not ready:
        print(f'Initializing storage: gs://{storage_bucket}{storage_path}')
        raise NotImplementedError('Storage initialization for new instances is not supported yet')


def delete(storage_path, stoarge_bucket=None):
    print('WARNING! Storage was not deleted')



from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.providers import manager as providers_manager
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager

from ckan_cloud_operator.providers.users.constants import PROVIDER_SUBMODULE as users_provider_submodule
from .constants import CRD_KIND, CRD_PLURAL, CRD_SINGULAR


def initialize(default_provider=None, interactive=False):
    crds_manager.install_crd(CRD_SINGULAR, CRD_PLURAL, CRD_KIND)
    if not default_provider:
        default_provider = 'gcloud' if cluster_manager.get_provider_id() == 'gcloud' else 'rancher'
    providers_manager.get_provider(users_provider_submodule, required=False, default=default_provider).initialize(interactive=interactive)


def create(name, role):
    logs.info(f'Creating user {name} (role={role})')
    try:
        extra_label_suffixes = get_user_labels(name, role)
    except AssertionError:
        logs.info('No user labels found, initializing and retrying')
        initialize()
        extra_label_suffixes = get_user_labels(name, role)

    kubectl.apply(crds_manager.get_resource(
        CRD_SINGULAR,
        name,
        spec=_get_spec(name, role),
        extra_label_suffixes=extra_label_suffixes
    ))


def update(name):
    logs.info(f'Updating user: {name}')
    providers_manager.get_provider(users_provider_submodule, verbose=True).update(get(name))


def get(name):
    return crds_manager.get(CRD_SINGULAR, name=name)


def get_kubeconfig(name):
    return _get_provider().get_kubeconfig(get(name))


def get_user_labels(name, role):
    return {'name': name, 'role': role}


def get_role_labels(role):
    return {'role': role}


def _get_spec(name, role):
    return {'name': name, 'role': role}


def _get_provider(verbose=False):
    return providers_manager.get_provider(users_provider_submodule, verbose=verbose)



import click
import traceback

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from . import manager


@click.group()
def env():
    '''Manage CKAN Environments
    '''
    pass

@env.command('list')
def list_env():
    '''
    Lists existing environments

    \b
    cco ckan env list
    > POC
    > DEV
    > PRD
    '''
    manager.list_environments()

@env.command()
@click.argument('ENVIRONMENT')
@click.option('--cloud-provider', help='One of minukube, azure, gcp, aws', required=True)
@click.option('--cluster-name', help='Kubernetes cluster name', required=True)
@click.option('--resource-group', help='Azure resource group name  [Azure only]')
@click.option('--subscription', help='Azure subscription id  [Azure only]')
@click.option('--region', help='GCP region id Eg: europe-west1  [GCP only]')
@click.option('--project', help='GCP project name  [GCP only]')
def add(environment, **kwargs):
    '''
    Adds an environment

    ENVIRONMENT: The name of the environment Eg: poc

    \b
    cco ckan env add poc --cloud-provider cloud-provider-name \\
                         --resource-group resource-group-name \\
                         --cluster-name cluster-name \\
                         --subscription subscription-id \\
                         --region region-name \\
                         --project project-name
    > POC environment was succefully added
    '''
    manager.add_environment(environment, **kwargs)

@env.command()
@click.argument('ENVIRONMENT')
@click.option('--cloud-provider', help='One of minukube, azure, gcp, aws')
@click.option('--cluster-name', help='Kubernetes cluster name')
@click.option('--resource-group', help='Azure resource group name  [Azure only]')
@click.option('--subscription', help='Azure subscription id  [Azure only]')
@click.option('--region', help='GCP region id Eg: europe-west1  [GCP only]')
@click.option('--project', help='GCP project name  [GCP only]')
def update(environment, **kwargs):
    '''
    Update configurations for given environment

    ENVIRONMENT: The name of the environment Eg: dev

    \b
    cco ckan env add poc --cloud-provider cloud-provider-name \\
                         --resource-group resource-group-name \\
                         --cluster-name new-cluster-name \\
                         --subscription subscription-id \\
                         --region region-name \\
                         --project project-name
    > POC environment was succefully updated
    '''
    manager.update_environment(environment, **kwargs)

@env.command()
@click.argument('ENVIRONMENT')
def set(environment):
    '''
    Sets given environment as a current working environment

    ENVIRONMENT: The name of the environment Eg: dev

    \b
    cco ckan env set poc
    > You are working with POC environment now
    '''
    manager.set_environment(environment)

@env.command()
@click.argument('ENVIRONMENT')
def rm(environment):
    '''
    Deletes given environment

    ENVIRONMENT: The name of the environment Eg: dev

    cco ckan env rm poc
    > POC environment was succesfully removed
    '''
    manager.remove_environment(environment)



import unittest
from unittest.mock import patch, MagicMock

from ckan_cloud_operator.routers.annotations import CkanRoutersAnnotations
from ckan_cloud_operator.routers.traefik import manager


class RoutersManagerTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.providers.cluster.manager.get_provider_id')
    @patch('ckan_cloud_operator.kubectl.apply')
    @patch.object(CkanRoutersAnnotations, 'update_flag')
    def test_create_with_cloudflare(self, update_flag, apply, get_provider_id):
        get_provider_id.return_value = 'gcp'
        router = {
            'metadata': {
                'name': 'datapusher',
            },
            'spec': {
                'default-root-domain': 'default',
                'type': 'traefik',
                'cloudflare': {
                    'email': 'admin@localhost',
                    'api-key': 'apikey'
                },
            }
        }
        manager.create(router)
        apply.assert_called_once_with(router)

    @patch('ckan_cloud_operator.providers.cluster.manager.get_provider_id')
    @patch.object(CkanRoutersAnnotations, 'update_flag')
    def test_create_with_cloudflare_without_credentials(self, update_flag, get_provider_id):
        get_provider_id.return_value = 'gcp'
        router = {
            'metadata': {
                'name': 'datapusher',
            },
            'spec': {
                'default-root-domain': None,
                'type': 'traefik',
                'dns-provider': 'cloudflare',
                'cloudflare': {
                },
            }
        }
        with self.assertRaisesRegex(AssertionError, 'invalid traefik router spec'):
            manager.create(router)

    @patch('ckan_cloud_operator.providers.cluster.manager.get_provider_id')
    @patch('ckan_cloud_operator.kubectl.apply')
    @patch.object(CkanRoutersAnnotations, 'update_flag')
    def test_create_with_route53(self, update_flag, apply, get_provider_id):
        get_provider_id.return_value = 'aws'
        router = {
            'metadata': {
                'name': 'datapusher',
            },
            'spec': {
                'default-root-domain': 'default',
                'type': 'traefik',
            }
        }
        manager.create(router)
        apply.assert_called_once_with(router)

    @patch('ckan_cloud_operator.kubectl.call')
    @patch('ckan_cloud_operator.kubectl.get_items_by_labels')
    @patch('ckan_cloud_operator.kubectl.get')
    def test_delete(self, get, get_items_by_labels, call):
        get_items_by_labels.return_value = []
        get.return_value = False
        call.return_value = 0
        manager.delete('datapushers')
        self.assertEqual(call.call_count, 7)
        get_items_by_labels.assert_called_once_with('CkanCloudRoute', {'ckan-cloud/router-name': 'datapushers'}, required=False)

    @patch('ckan_cloud_operator.kubectl.call')
    def test_delete_failed(self, call):
        call.return_value = 1
        with self.assertRaisesRegex(Exception, 'Deletion failed'):
            manager.delete('datapushers')
        self.assertEqual(call.call_count, 7)

    @patch('ckan_cloud_operator.routers.traefik.manager.traefik_deployment')
    def test_get_with_default_params(self, traefik_deployment):
        traefik_deployment.get_dns_data.return_value = {'root-domain': 'ckan.io'}
        traefik_deployment.get.return_value = {'router-type': 'traefik'}
        self.assertEqual(manager.get('datapushers'), {'router-type': 'traefik'})

    @patch('ckan_cloud_operator.routers.traefik.manager.traefik_deployment')
    def test_get_dns_attr(self, traefik_deployment):
        traefik_deployment.get_dns_data.return_value = {'root-domain': 'ckan.io'}
        traefik_deployment.get.return_value = {'router-type': 'traefik'}
        self.assertEqual(manager.get('datapushers', attr='dns'), {'root-domain': 'ckan.io'})

    @patch('ckan_cloud_operator.routers.traefik.manager.traefik_deployment')
    def test_get_all_attr(self, traefik_deployment):
        traefik_deployment.get_dns_data.return_value = {'root-domain': 'ckan.io'}
        traefik_deployment.get.return_value = {'router-type': 'traefik'}
        self.assertEqual(manager.get('datapushers', attr='all'), {'dns': {'root-domain': 'ckan.io'}, 'deployment': {'router-type': 'traefik'}})

    @patch('ckan_cloud_operator.kubectl.get')
    def test_update_new_deployment(self, get):
        get.return_value = None
        annotations = MagicMock()

        manager.update('datapushers', False, {'router-type': 'traefik'}, annotations, {'root-domain': 'ckan.io'})
        get.assert_called_once_with('deployment router-traefik-datapushers', required=False)
        self.assertEqual(annotations.update_status.call_count, 1)

    @patch('ckan_cloud_operator.kubectl.get')
    def test_update_old_deployment(self, get):
        get.side_effect = [
            {
                'metadata': {
                    'generation': 1
                }
            },
            {
                'metadata': {
                    'generation': 2
                }
            }
        ]
        annotations = MagicMock()

        manager.update('datapushers', False, {'router-type': 'traefik'}, annotations, {'root-domain': 'ckan.io'})
        self.assertEqual(get.call_count, 2)
        self.assertEqual(annotations.update_status.call_count, 1)

    @patch('ckan_cloud_operator.kubectl.get')
    def test_update_old_deployment_wrong_generation(self, get):
        get.side_effect = [
            {
                'metadata': {
                    'generation': 1
                }
            },
            {
                'metadata': {
                    'generation': 4
                }
            }
        ]
        annotations = MagicMock()

        with self.assertRaisesRegex(Exception, 'Invalid generation: 4 \(expected: 2\)'):
            manager.update('datapushers', False, {'router-type': 'traefik'}, annotations, {'root-domain': 'ckan.io'})
        self.assertEqual(get.call_count, 2)
        self.assertEqual(annotations.update_status.call_count, 1)



import yaml
from ckan_cloud_operator import kubectl
from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator.config import manager as config_manager

def add_cli_commands(click, command_group, great_success):

    @command_group.command('initialize')
    def datapushers_initialize():
        initialize()
        great_success()

    @command_group.command('create')
    @click.argument('DATAPUSHER_NAME')
    @click.argument('DOCKER_IMAGE')
    @click.argument('PATH_TO_CONFIG_YAML')
    def datapushers_create(datapusher_name, docker_image, path_to_config_yaml):
        """Create and update a DataPusher deployment

        Example:

            ckan-cloud-operator datapushers create datapusher-1 registry.gitlab.com/viderum/docker-datapusher:cloud-datapusher-1-v9 /path/to/datapusher-1.yaml
        """
        create(datapusher_name, docker_image, path_to_config_yaml)
        update(datapusher_name)
        great_success()

    @command_group.command('update')
    @click.argument('DATAPUSHER_NAME')
    def datapushers_update(datapusher_name):
        update(datapusher_name)
        great_success()

    @command_group.command('list')
    @click.option('--full', is_flag=True)
    def datapushers_list(full):
        print(yaml.dump(list(full=full), default_flow_style=False))

    @command_group.command('get')
    @click.argument('DATAPUSHER_NAME')
    def datapushers_get(datapusher_name):
        print(yaml.dump(get(datapusher_name), default_flow_style=False))

    @command_group.command('delete')
    @click.argument('DATAPUSHER_NAME')
    def datapushers_delete(datapusher_name):
        delete(datapusher_name)
        great_success()


def initialize():
    install_crds()
    datapusher_envvars = {'PORT': '8000'}
    router_name = 'datapushers'
    if not routers_manager.get(router_name, required=False):
        routers_manager.create(router_name, routers_manager.get_traefik_router_spec())
    if config_manager.get('enable-deis-ckan', configmap_name='global-ckan-config') == 'y':
        create(
            'datapusher-1',
            'registry.gitlab.com/viderum/docker-datapusher:cloud-datapusher-1-v9',
            datapusher_envvars,
            router_name
        )
        create(
            'datapusher-de',
            'registry.gitlab.com/viderum/docker-datapusher:cloud-de-git-943fc3e0',
            datapusher_envvars,
            router_name
        )
        create(
            'datapusher-giga',
            'registry.gitlab.com/viderum/docker-datapusher:cloud-giga-git-2b05b22d',
            datapusher_envvars,
            router_name
        )
        create(
            'datapusher-increased-max-length',
            'registry.gitlab.com/viderum/docker-datapusher:cloud-increased-max-length-git-84e86116',
            datapusher_envvars,
            router_name
        )
        update('datapusher-1')
        update('datapusher-de')
        update('datapusher-giga')
        update('datapusher-increased-max-length')
        routers_manager.update(router_name)


def install_crds():
    """Ensures installaion of the datapusher custom resource definitions on the cluster"""
    kubectl.install_crd('ckanclouddatapushers', 'ckanclouddatapusher', 'CkanCloudDatapusher')


def create(name, image, config, router_name=None):
    labels =  _get_labels(name)
    datapusher = kubectl.get_resource('stable.viderum.com/v1', 'CkanCloudDatapusher', name, labels)
    datapusher['spec'] = {'image': image,
                          'config': config}
    kubectl.apply(datapusher)
    if router_name:
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'datapusher',
            'datapusher-name': name
        })


def update(name):
    _update_registry_secret()
    datapusher = kubectl.get(f'CkanCloudDatapusher {name}')
    deployment_name = get_deployment_name(name)
    labels = _get_labels(name)
    spec = _get_deployment_spec(labels, datapusher['spec'])
    print(f'Updating CkanCloudDatapusher {name} (deployment_name={deployment_name})')
    deployment = kubectl.get_deployment(deployment_name, labels, spec)
    kubectl.apply(deployment)


def delete(name):
    print(f'Deleting datapusher {name}')
    assert kubectl.remove_resource_and_dependencies(
        'CkanCloudDatapusher',
        name,
        ['deployment', 'service'],
        f'ckan-cloud/datapusher-name={name}'
    )


def update_service(name, labels):
    labels.update(**_get_labels(name))
    service = kubectl.get_resource('v1', 'Service', get_service_name(name), labels)
    service['spec'] = {
        'ports': [
            {'name': '8000', 'port': 8000}
        ],
        'selector': {
            'ckan-cloud/datapusher-name': name,
            'app': 'datapusher',
        }
    }
    kubectl.apply(service)


def get_service_url(name):
    service_name = get_service_name(name)
    namespace = 'ckan-cloud'
    port = 8000
    return f'http://{service_name}.{namespace}:{port}'


def get(name, deployment=None, full=False):
    deployment_name = get_deployment_name(name)
    if not deployment:
        deployment = kubectl.get(f'deployment {deployment_name}', required=False)
    if deployment:
        deployment_status = kubectl.get_deployment_detailed_status(
            deployment,
            f'ckan-cloud/datapusher-name={name}',
            'datapusher'
        )
        ready = bool(len(deployment_status.get('errors', [])) == 0)
    else:
        deployment_status = {}
        ready = False
    if full:
        return {
            'ready': ready,
            'name': name,
            **deployment_status
        }
    else:
        return {
            'ready': ready,
            'name': name
        }


def list(full=False):
    return [get(datapusher['metadata']['name'], full=full) for datapusher in kubectl.get('CkanCloudDatapusher')['items']]


def _get_labels(name):
    return {'ckan-cloud/datapusher-name': name}


def get_deployment_name(name):
    return name if name.startswith('datapusher') else f'datapusher-{name}'


def get_service_name(name):
    return get_deployment_name(name)


def _get_deployment_spec(labels, spec):
    deployment_spec = {
        'replicas': 1,
        'revisionHistoryLimit': 10,
        'template': {
            'metadata': {
                'labels': dict(labels, app='datapusher')
            },
            'spec': {
                'imagePullSecrets': [
                    {'name': 'datapushers-docker-registry'}
                ],
                'containers': [
                    {
                        'name': 'datapusher',
                        'image': spec['image'],
                        'env': [
                            *_get_deployment_pod_envvars(spec['config']),
                        ],
                        'readinessProbe': {
                            'failureThreshold': 1,
                            'initialDelaySeconds': 5,
                            'periodSeconds': 5,
                            'successThreshold': 1,
                            'tcpSocket': {
                                'port': 8000
                            },
                            'timeoutSeconds': 5
                        }
                    }
                ]
            }
        }
    }
    kubectl.add_operator_timestamp_annotation(deployment_spec['template']['metadata'])
    return deployment_spec


def _get_deployment_pod_envvars(config):
    return [{'name': k, 'value': v} for k, v in config.items()]


def _update_registry_secret():
    secret = kubectl.get('secret datapushers-docker-registry', required=False)
    if secret:
        print('Secret already exists, delete to recreate: datapushers-registry-secret')
    else:
        print('Creating datapushers registry secret')
        from ckan_cloud_operator.providers.ckan import manager as ckan_manager
        docker_server, docker_username, docker_password, docker_email = ckan_manager.get_docker_credentials()
        kubectl.check_call(f'create secret docker-registry datapushers-docker-registry '
                           f'--docker-password={docker_password} '
                           f'--docker-server={docker_server} '
                           f'--docker-username={docker_username} '
                           f'--docker-email={docker_email}')



import binascii
import os
import datetime
import traceback
import time

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from ckan_cloud_operator.drivers.postgres import driver as postgres_driver

from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.providers.db.proxy import manager as db_proxy_manager
from ckan_cloud_operator.providers.db import manager as db_manager
from ckan_cloud_operator.providers import manager as providers_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager

from ckan_cloud_operator.infra import CkanInfra

from ckan_cloud_operator.providers.cluster.constants import PROVIDER_SUBMODULE as cluster_provider_submodule
from ckan_cloud_operator.providers.db.constants import PROVIDER_SUBMODULE as db_provider_submodule

from .constants import PROVIDER_SUBMODULE

from .constants import (
    MIGRATION_CRD_KIND as CRD_KIND,
    MIGRATION_CRD_PLURAL as CRD_PLURAL,
    MIGRATION_CRD_SINGULAR as CRD_SINGULAR,
    MIGRATION_EVENT_STATUS_SUCCESS as STATUS_SUCCESS,
    MIGRATION_EVENT_STATUS_FAILURE as STATUS_FAILURE
)


def initialize(log_kwargs=None, interactive=False):
    log_kwargs = log_kwargs or {}
    logs.info(f'Installing crds', **log_kwargs)
    crds_manager.install_crd(CRD_SINGULAR, CRD_PLURAL, CRD_KIND, hash_names=True)
    if cluster_manager.get_provider_id() == 'gcloud':
        ckan_infra = CkanInfra(required=False)
        if interactive:
            providers_manager.config_interactive_set(
                PROVIDER_SUBMODULE,
                default_values={'gcloud-storage-import-bucket': ckan_infra.GCLOUD_SQL_DEIS_IMPORT_BUCKET},
                suffix='deis-migration'
            )
        else:
            if not providers_manager.config_get(PROVIDER_SUBMODULE, key='gcloud-storage-import-bucket', suffix='deis-migration', required=False):
                providers_manager.config_set(
                    PROVIDER_SUBMODULE,
                    values={'gcloud-storage-import-bucket': ckan_infra.GCLOUD_SQL_DEIS_IMPORT_BUCKET},
                    suffix='deis-migration'
                )


def get(name=None, required=True):
    return crds_manager.get(CRD_SINGULAR, name=name, required=required)


def create(name, spec, force=False, exists_ok=False, delete_dbs=False):
    migration = get(name, required=False)
    if migration:
        if force or delete_dbs:
            delete(name, delete_dbs=delete_dbs)
        elif exists_ok:
            return migration
        else:
            logs.error('migration name conflict: ' + str(migration.get('spec')))
            raise Exception(f'migration already exists: {name}, Run with --force / --rerun / --recreate-dbs')
    logs.info(f'Creating migration name {name} (spec={spec})')
    migration = crds_manager.get_resource(
        CRD_SINGULAR,
        name,
        spec=_get_spec(name, spec),
        extra_label_suffixes=_get_labels(name, spec)
    )
    kubectl.apply(migration)
    return migration


def delete(name, delete_dbs=False):
    migration = crds_manager.get(CRD_SINGULAR, name=name, required=False) or {}
    if delete_dbs:
        db_prefix = migration.get('spec', {}).get('db-prefix') or ''
        admin_connection_string = db_manager.get_external_admin_connection_string(db_prefix=db_prefix)
        db_name = migration.get('spec', {}).get('datastore-name')
        datastore_name = migration.get('spec', {}).get('db-name')
        datastore_ro_name = crds_manager.config_get(CRD_SINGULAR, name, key='datastore-readonly-user-name', is_secret=True, required=False)
        if db_name or datastore_name or datastore_ro_name:
            with postgres_driver.connect(admin_connection_string) as admin_conn:
                _delete_dbs(admin_conn, db_name, datastore_name, datastore_ro_name)
    crds_manager.delete(CRD_SINGULAR, name)


def migrate_deis_dbs(old_site_id=None, db_name=None, datastore_name=None, force=False, rerun=False, recreate_dbs=False,
                     dbs_suffix=None, skip_create_dbs=False, skip_datastore_import=False,
                     db_import_url=None, datastore_import_url=None, db_prefix=None):
    if not dbs_suffix: dbs_suffix = ''
    if not db_prefix: db_prefix = db_manager.get_default_db_prefix()
    if db_import_url or datastore_import_url:
        assert db_import_url and datastore_import_url
        assert db_name and datastore_name
        assert not old_site_id
        logs.info(f'Restoring from backup: {db_import_url}, {datastore_import_url} -> {db_name}, {datastore_name} ({db_prefix})')
        assert not skip_datastore_import
        migration_name = f'restore-{db_name}-{datastore_name}'
        if len(migration_name) > 50:
            migration_name = f'rr-{db_name}'
        if db_prefix:
            migration_name += f'-{db_prefix}'
    else:
        assert old_site_id
        assert not db_import_url and not datastore_import_url
        if not db_name: db_name = f'{old_site_id}{dbs_suffix}'
        if not datastore_name: datastore_name = f'{db_name}{dbs_suffix}-datastore'
        logs.info(f'Starting migration ({old_site_id} -> {db_name}, {datastore_name})')
        if skip_datastore_import:
            logs.warning('skipping datastore DB import')
        migration_name = f'deis-dbs-{old_site_id}-to-{db_name}--{datastore_name}'
        if len(migration_name) > 50:
            migration_name = f'dd-{old_site_id}-{db_name}'
        if db_prefix:
            migration_name += f'-{db_prefix}'
    migration = create(
        name=migration_name,
        spec={
            'type': 'deis-ckan',
            'old-site-id': old_site_id,
            'db-name': db_name,
            'datastore-name': datastore_name,
            'skip-datastore-import': skip_datastore_import,
            'db-import-url': db_import_url,
            'datastore-import-url': datastore_import_url,
            'db-prefix': db_prefix
        },
        force=force,
        exists_ok=rerun and not force,
        delete_dbs=recreate_dbs
    )
    yield {'step': 'created-migration-resource',
           'msg': f'Created the migration custom resource: {migration_name}',
           'migration-name': migration_name}
    yield from update(migration, recreate_dbs=recreate_dbs, skip_create_dbs=skip_create_dbs)
    yield {'step': 'migration-complete', 'msg': f'Migration completed successfully for: {migration_name}', 'status': STATUS_SUCCESS}


def update(migration, recreate_dbs=False, skip_create_dbs=False):
    migration_type = migration['spec']['type']
    if migration_type == 'deis-ckan':
        if migration['spec'].get('imported-data'):
            yield {'step': 'imported-data', 'msg': 'data already imported'}
        else:
            migration_name = migration['spec']['name']
            db_name = migration['spec']['db-name']
            datastore_name = migration['spec']['datastore-name']
            datastore_ro_name = _get_or_create_datastore_readonly_user_name(migration_name, datastore_name)
            db_prefix = migration['spec'].get('db-prefix')
            if not skip_create_dbs:
                yield from _create_base_dbs_and_roles(migration_name, db_name, datastore_name, recreate_dbs, datastore_ro_name,
                                                      db_prefix=db_prefix)
                yield from _initialize_postgis_extensions(db_name, db_prefix)
            yield from _import_data(migration['spec']['old-site-id'], db_name, datastore_name,
                                    skip_datastore_import=migration['spec'].get('skip-datastore-import'),
                                    db_import_url=migration['spec'].get('db-import-url'),
                                    datastore_import_url=migration['spec'].get('datastore-import-url'),
                                    db_prefix=db_prefix)
            migration['spec']['imported-data'] = True
            kubectl.apply(migration)
    elif migration_type == 'new-db':
        if migration['spec'].get('created-db'):
            yield {'step': 'created-db', 'msg': 'DB Already created'}
        else:
            migration_name = migration['spec']['name']
            db_name = migration['spec']['db-name']
            db_prefix = migration['spec'].get('db-prefix')
            yield from _create_base_dbs_and_roles(migration_name, db_name, None, recreate_dbs, None, db_prefix=db_prefix)
            yield from _initialize_postgis_extensions(db_name, db_prefix)
            migration['spec']['created-db'] = True
            kubectl.apply(migration)
    elif migration_type == 'new-datastore':
        if migration['spec'].get('created-datastore'):
            yield {'step': 'created-datastore', 'msg': 'Datastore Already created'}
        else:
            migration_name = migration['spec']['name']
            datastore_name = migration['spec']['datastore-name']
            datastore_ro_name = _get_or_create_datastore_readonly_user_name(migration_name, datastore_name)
            db_prefix = migration['spec'].get('db-prefix')
            yield from _create_base_dbs_and_roles(migration_name, None, datastore_name, recreate_dbs, datastore_ro_name, db_prefix=db_prefix)
            migration['spec']['created-datastore'] = True
            kubectl.apply(migration)
    else:
        raise Exception(f'Unknown migration type: {migration_type}')


def get_all_dbs_users():
    dbs, users = [], []
    for migration in crds_manager.get(CRD_SINGULAR)['items']:
        migration_name = migration['spec']['name']
        spec = migration['spec']
        if spec.get('type') == 'deis-ckan':
            db_name = spec['db-name']
            datastore_ro_name = get_datastore_raedonly_user_name(migration_name, required=False)
            datastore_name = spec['datastore-name']
            db_password, datastore_password, datastore_ro_password = get_dbs_passwords(migration_name, required=False)
            if all([db_password, datastore_password, datastore_ro_password, db_name, datastore_name, datastore_ro_name]):
                db_host, db_port = db_manager.get_internal_unproxied_db_host_port()
                dbs.append((db_name, db_host, db_port))
                dbs.append((datastore_name, db_host, db_port))
                users.append((db_name, db_password))
                users.append((datastore_name, datastore_password))
                users.append((datastore_ro_name, datastore_ro_password))
        elif spec.get('type') == 'new-db':
            db_name = spec['db-name']
            db_password, _, _ = get_dbs_passwords(migration_name, required=False)
            if db_password:
                db_host, db_port = db_manager.get_internal_unproxied_db_host_port()
                dbs.append((db_name, db_host, db_port))
                users.append((db_name, db_password))
        elif spec.get('type') == 'new-datastore':
            datastore_ro_name = get_datastore_raedonly_user_name(migration_name, required=False)
            datastore_name = spec['datastore-name']
            _, datastore_password, datastore_ro_password = get_dbs_passwords(migration_name, required=False)
            if datastore_password and datastore_ro_password:
                db_host, db_port = db_manager.get_internal_unproxied_db_host_port()
                dbs.append((datastore_name, db_host, db_port))
                users.append((datastore_name, datastore_password))
                users.append((datastore_ro_name, datastore_ro_password))
    return dbs, users


def print_event_exit_on_complete(event, details_msg, soft_exit=False):
    if not details_msg: details_msg = ''
    msg = event.pop('msg')
    step = event.pop('step')
    logs.info(f'{step}: {msg}')
    if len(event)> 0:
        logs.info(f'event metadata: {event}')
    event_completed_successfully = get_event_migration_complete_status(event)
    if event_completed_successfully:
        logs.info(f'Migration completed successfully ({details_msg})')
        if soft_exit:
            return True
        else:
            logs.exit_great_success()
    elif event_completed_successfully is False:
        logs.error(f'Migration failed ({details_msg})')
        if soft_exit:
            return False
        else:
            logs.exit_catastrophic_failure()
    else:
        return None


def get_event_migration_complete_status(event):
    status = event.get('status')
    if not status: return None
    return {
        STATUS_SUCCESS: True,
        STATUS_FAILURE: False,
    }.get(status, None)


def get_event_migration_created_name(event):
    if event.get('step') == 'created-migration-resource':
        return event['migration-name']
    else:
        return None


def get_dbs_passwords(migration_name, required=False):
    database_password, datastore_password, datastore_readonly_password = [
        crds_manager.config_get(
            CRD_SINGULAR, migration_name, key=key, is_secret=True, required=required
        )
        for key in ['database-password', 'datastore-password', 'datastore-readonly-password']
    ]
    return database_password, datastore_password, datastore_readonly_password


def get_datastore_raedonly_user_name(migration_name, required=False):
    return crds_manager.config_get(CRD_SINGULAR, migration_name,
                                   key='datastore-readonly-user-name', is_secret=True, required=required)


def get_db_import_urls(old_site_id):
    import_backup = providers_manager.config_get(PROVIDER_SUBMODULE, key='gcloud-storage-import-bucket', suffix='deis-migration', required=True)
    instance_latest_datestring = None
    instance_latest_dt = None
    instance_latest_datastore_datestring = None
    instance_latest_datastore_dt = None
    for line in _gcloud().check_output(f"ls 'gs://{import_backup}/postgres/????????/*.sql'",
                                       gsutil=True).decode().splitlines():
        # gs://viderum-deis-backups/postgres/20190122/nav.20190122.dump.sql
        datestring, filename = line.split('/')[4:]
        file_instance = '.'.join(filename.split('.')[:-3])
        is_datastore = file_instance.endswith('-datastore')
        file_instance = file_instance.replace('-datastore', '')
        dt = datetime.datetime.strptime(datestring, '%Y%m%d')
        if file_instance == old_site_id:
            if is_datastore:
                if instance_latest_datastore_dt is None or instance_latest_datastore_dt < dt:
                    instance_latest_datastore_datestring = datestring
                    instance_latest_datastore_dt = dt
            elif instance_latest_dt is None or instance_latest_dt < dt:
                instance_latest_datestring = datestring
                instance_latest_dt = dt
    return (
        f'gs://{import_backup}/postgres/{instance_latest_datestring}/{old_site_id}.{instance_latest_datestring}.dump.sql' if instance_latest_datestring else None,
        f'gs://{import_backup}/postgres/{instance_latest_datastore_datestring}/{old_site_id}-datastore.{instance_latest_datastore_datestring}.dump.sql' if instance_latest_datastore_datestring else None
    )


def _get_spec(name, spec):
    return dict(name=name, **spec)


def _get_labels(name, spec):
    labels = {'name': name}
    if 'old-site-id' in spec:
        labels['old-site-id'] = spec['old-site-id']
    return labels


def _get_or_create_migration_db_passwords(migration_name, create_if_not_exists=True, skip_keys=None):
    passwords = []
    errors = []
    for password_config_key in ['database-password', 'datastore-password', 'datastore-readonly-password']:
        if skip_keys is not None and password_config_key not in skip_keys:
            passwords.append(None)
            continue
        password = crds_manager.config_get(
            CRD_SINGULAR, migration_name, key=password_config_key, is_secret=True, required=False
        )
        if create_if_not_exists:
            if password:
                logs.info(f'Password already exists: {password_config_key}')
                errors.append(f'password-exists: {password_config_key}')
            else:
                logs.info(f'Generating new password for {password_config_key}')
                password = _generate_password()
                crds_manager.config_set(
                    CRD_SINGULAR, migration_name,
                    key=password_config_key, value=password,
                    is_secret=True
                )
        passwords.append(password)
    return [errors] + passwords


def _generate_password():
    return binascii.hexlify(os.urandom(12)).decode()


def _get_or_create_datastore_readonly_user_name(migration_name, datastore_name):
    name = crds_manager.config_get(CRD_SINGULAR, migration_name,
                                   key='datastore-readonly-user-name', is_secret=True, required=False)
    if not name:
        name = f'{datastore_name}-ro'
        crds_manager.config_set(CRD_SINGULAR, migration_name,
                                key='datastore-readonly-user-name', value=name, is_secret=True)
    return name


def _initialize_postgis_extensions(db_name, db_prefix):
    logs.info('initializing postgis extensions for main db')
    connection_string = db_manager.get_external_admin_connection_string(db_name, db_prefix=db_prefix)
    with postgres_driver.connect(connection_string) as conn:
        postgres_driver.initialize_extensions(conn, [
            'postgis', 'postgis_topology', 'fuzzystrmatch', 'postgis_tiger_geocoder'
        ])
        with conn.cursor() as cur:
            try:
                cur.execute(f'ALTER TABLE spatial_ref_sys OWNER TO "{db_name}"')
            except Exception:
                traceback.print_exc()
    yield {'step': 'initialize-postgis', 'msg': f'Initialized postgis for db: {db_name} ({db_prefix})'}


def _update_db_proxy(db_name, datastore_name, datastore_ro_name, db_password, datastore_password, datastore_ro_password, db_prefix):
    logs.info('Updating db proxy')
    db_proxy_manager.update(wait_updated=False)
    ok = False
    for i in range(5):
        try:
            for user, password, db in [(db_name, db_password, db_name),
                                       (datastore_name, datastore_password, datastore_name),
                                       (datastore_ro_name, datastore_ro_password, datastore_name)]:
                if user:
                    db_manager.check_connection_string(
                        db_manager.get_external_connection_string(user, password, db, db_prefix=db_prefix)
                    )
            ok = True
            break
        except Exception as e:
            logs.warning(str(e))
        logs.info(f'Waiting for connection to db proxy...')
        # 40 seconds on first iteration - to ensure secret is updated in pgbouncer volume
        time.sleep(40 if i == 0 else 1)
        db_proxy_manager.reload()
        time.sleep(10 if i == 2 else 5)
    assert ok, 'failed to get connection to db proxy'
    yield {'step': 'update-db-proxy',
           'msg': f'Updated DB Proxy with the new dbs and roles: {db_name}, {datastore_name}, {datastore_ro_name} ({db_prefix})'}


def _create_base_dbs_and_roles(migration_name, db_name, datastore_name, recreate_dbs, datastore_ro_name, db_prefix=None):
    logs.info('Creating base DBS')
    admin_connection_string = db_manager.get_external_admin_connection_string(db_prefix=db_prefix)
    with postgres_driver.connect(admin_connection_string) as admin_conn:
        if recreate_dbs:
            _delete_dbs(admin_conn, db_name, datastore_name, datastore_ro_name)
        if not datastore_name:
            skip_keys = ['database-password']
        elif not db_name:
            skip_keys = ['datastore-password', 'datastore-readonly-password']
        else:
            skip_keys = None
        password_errors, db_password, datastore_password, datastore_ro_password = _get_or_create_migration_db_passwords(
            migration_name,
            skip_keys=skip_keys
        )
        yield {'step': 'get-create-passwords', 'msg': 'Created Passwords'}
        admin_user = db_manager.get_admin_db_user(db_prefix=db_prefix)
        if db_name:
            db_errors = postgres_driver.create_base_db(admin_conn, db_name, db_password, grant_to_user=admin_user)
        else:
            db_errors = []
        if datastore_name:
            datastore_errors = postgres_driver.create_base_db(admin_conn, datastore_name, datastore_password,
                                                              grant_to_user=admin_user)
        else:
            datastore_errors = []
        if datastore_name and db_name:
            assert (len(datastore_errors) == 0 and len(db_errors) == 0) or len(password_errors) == 3, \
                'some passwords were not created, but DB / roles need to be created, we cannot know the right passwords'
        yield {'step': 'create-base-dbs', 'msg': f'Created Base DB and roles: {db_name}, {datastore_name} ({db_prefix})'}
        if datastore_name:
            postgres_driver.create_role_if_not_exists(admin_conn, datastore_ro_name, datastore_ro_password)
            yield {'step': 'created-datastore-ro-role', 'msg': f'Created Datastore read-only user: {datastore_ro_name} ({db_prefix})'}
    yield {'step': 'created-base-dbs-and-roles', f'msg': f'Created base dbs and roles: {db_name}, {datastore_name}, {datastore_ro_name} ({db_prefix})'}
    yield from _update_db_proxy(db_name, datastore_name, datastore_ro_name, db_password, datastore_password, datastore_ro_password, db_prefix)


def _delete_dbs(admin_conn, db_name, datastore_name, datastore_ro_name):
    if db_name:
        postgres_driver.delete_base_db(admin_conn, db_name)
    if datastore_name:
        postgres_driver.delete_base_db(admin_conn, datastore_name)
    if datastore_ro_name:
        postgres_driver.delete_role(admin_conn, datastore_ro_name)


def _import_data(old_site_id, db_name, datastore_name, skip_datastore_import=False,
                 db_import_url=None, datastore_import_url=None,
                 import_user=None, db_prefix=None):
    if db_import_url or datastore_import_url:
        assert db_import_url and datastore_import_url
        db_url, datastore_url = db_import_url, datastore_import_url
    else:
        db_url, datastore_url = get_db_import_urls(old_site_id)
    assert db_url and (datastore_url or skip_datastore_import), f'failed to find db import urls for old site id {old_site_id}'
    _gcloudsql().import_db(db_url, db_name, import_user=import_user or db_name, db_prefix=db_prefix)
    yield {'step': 'import-db-data', 'msg': f'Imported DB: {db_name}'}
    if skip_datastore_import:
        yield {'step': 'import-datastore-data', 'msg': 'skipped'}
    else:
        _gcloudsql().import_db(datastore_url, datastore_name, import_user=import_user or datastore_name, db_prefix=db_prefix)
        yield {'step': 'import-datastore-data', 'msg': f'Imported Datastore: {datastore_name}'}


def _gcloud():
    return providers_manager.get_provider(cluster_provider_submodule)


def _gcloudsql():
    return db_manager.get_provider()



import subprocess
import yaml
from ckan_cloud_operator import kubectl


class DeisCkanInstanceRegistry(object):

    def __init__(self, instance):
        self.instance = instance

    def update(self):
        self.instance.annotations.update_status('registry', 'created', lambda: self._create_instance_registry_secret())

    def delete(self):
        print('Deleting instance registry secret')
        subprocess.check_call(f'kubectl -n {self.instance.id} delete secret/{self.instance.id}-registry', shell=True)

    def get(self):
        exitcode, output = subprocess.getstatusoutput(f'kubectl -n {self.instance.id} get secret/{self.instance.id}-registry -o yaml')
        if exitcode == 0:
            return {'ready': len(kubectl.decode_secret(yaml.load(output))) > 0}
        else:
            return {'ready': False, 'error': output}

    def _create_instance_registry_secret(self):
        print('Creating instance registry secret')
        from ckan_cloud_operator.providers.ckan import manager as ckan_manager
        server, user, password, email = ckan_manager.get_docker_credentials()
        subprocess.call(f'kubectl -n {self.instance.id} create secret docker-registry {self.instance.id}-registry '
                              f'--docker-password={password} '
                              f'--docker-server={server} '
                              f'--docker-username={user} '
                              f'--docker-email={email}', shell=True)



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False, interactive=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file, interactive)

################################
# custom provider code starts here
#

import os
import binascii
import yaml
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.routers import manager as routers_manager


def initialize(interactive=False, storage_suffix=None, use_existing_disk_name=None, dry_run=False):
    _config_interactive_set({
        'disk-size-gb': None,
        **({} if storage_suffix else {'router-name': routers_manager.get_default_infra_router_name()})
    }, interactive=interactive, suffix=storage_suffix)
    _apply_secret(storage_suffix=storage_suffix)
    _apply_deployment(
        _get_or_create_volume(
            storage_suffix=storage_suffix,
            use_existing_disk_name=use_existing_disk_name
        ),
        storage_suffix=storage_suffix,
        dry_run=dry_run
    )
    _apply_service(storage_suffix=storage_suffix, dry_run=dry_run)
    if not storage_suffix:
        _update_route(storage_suffix=storage_suffix, dry_run=dry_run)
        _set_provider()


def print_credentials(raw=False, storage_suffix=None):
    hostname, access_key, secret_key = get_credentials(storage_suffix=storage_suffix)
    if raw:
        print(f'https://{hostname} {access_key} {secret_key}')
    else:
        print('Minio admin credentials:')
        print('External Domain: ' + hostname)
        print('Access Key: ' + access_key)
        print('Secret Key: ' + secret_key)
        print('\nto use with minio-client, run the following command:')
        print(f'mc config host add my-storage https://{hostname} {access_key} {secret_key}')



def get_credentials(storage_suffix=None):
    return [_get_frontend_hostname(storage_suffix=storage_suffix)] + [
        _config_get(key, required=True, is_secret=True, suffix=storage_suffix)
        for key in ['MINIO_ACCESS_KEY', 'MINIO_SECRET_KEY']
    ]


def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()


def _apply_secret(storage_suffix=None):
    access_key = _config_get('MINIO_ACCESS_KEY', required=False, is_secret=True, suffix=storage_suffix) or _generate_password(8)
    secret_key = _config_get('MINIO_SECRET_KEY', required=False, is_secret=True, suffix=storage_suffix) or _generate_password(12)
    _config_set(values={'MINIO_ACCESS_KEY': access_key, 'MINIO_SECRET_KEY': secret_key}, is_secret=True, suffix=storage_suffix)


def _apply_deployment(volume_spec, storage_suffix=None, dry_run=False):
    node_selector = volume_spec.pop('nodeSelector', None)
    if node_selector:
        pod_scheduling = {'nodeSelector': node_selector}
    else:
        pod_scheduling = {}
    container_spec_overrides = _config_get('container-spec-overrides', required=False, default=None, suffix=storage_suffix)
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(suffix=storage_suffix),
        _get_resource_labels(for_deployment=True, suffix=storage_suffix),
        {
            'replicas': 1,
            'revisionHistoryLimit': 10,
            'strategy': {'type': 'Recreate', },
            'selector': {
                'matchLabels': _get_resource_labels(for_deployment=True, suffix=storage_suffix)
            },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix=storage_suffix),
                    'annotations': _get_resource_annotations(suffix=storage_suffix)
                },
                'spec': {
                    **pod_scheduling,
                    'containers': [
                        {
                            'name': 'minio',
                            'image': 'minio/minio',
                            'args': ['server', '/export'],
                            'envFrom': [{'secretRef': {'name': _get_resource_name(suffix=storage_suffix)}}],
                            'ports': [{'containerPort': 9000}],
                            'volumeMounts': [
                                {
                                    'name': 'minio-data',
                                    'mountPath': '/export',
                                }
                            ],
                            **(json.loads(container_spec_overrides) if container_spec_overrides else {})
                        }
                    ],
                    'volumes': [
                        dict(volume_spec, name='minio-data')
                    ]
                }
            }
        }
    ), dry_run=dry_run)


def _apply_service(storage_suffix=None, dry_run=False):
    kubectl.apply(kubectl.get_resource(
        'v1', 'Service',
        _get_resource_name(suffix=storage_suffix),
        _get_resource_labels(suffix=storage_suffix),
        spec={
            'ports': [
                {'name': '9000', 'port': 9000}
            ],
            'selector': {
                'app': _get_resource_labels(for_deployment=True, suffix=storage_suffix)['app']
            }
        }
    ), dry_run=dry_run)


def _get_or_create_volume(storage_suffix=None, use_existing_disk_name=None, zone=0):
    disk_size_gb = _config_get('disk-size-gb', required=True, suffix=storage_suffix)
    volume_spec = _config_get('volume-spec', required=False, suffix=storage_suffix)
    if volume_spec:
        volume_spec = yaml.load(volume_spec)
    else:
        from ckan_cloud_operator.providers.cluster import manager as cluster_manager
        volume_spec = cluster_manager.create_volume(
            disk_size_gb,
            _get_resource_labels(suffix=storage_suffix),
            use_existing_disk_name=use_existing_disk_name, zone=zone
        )
        _config_set('volume-spec', yaml.dump(volume_spec, default_flow_style=False), suffix=storage_suffix)
    return volume_spec


def _update_route(storage_suffix=None, dry_run=False):
    backend_url_target_id = _get_backend_url_target_id(storage_suffix=storage_suffix)
    router_name = _config_get('router-name', required=True, suffix=storage_suffix)
    if not routers_manager.get_backend_url_routes(backend_url_target_id):
        deployment_name = _get_resource_name(suffix=storage_suffix)
        namespace = _get_namespace()
        subdomain_route = {
            'target-type': 'backend-url',
            'target-resource-id': backend_url_target_id,
            'backend-url': f'http://{deployment_name}.{namespace}:9000',
        }
        if dry_run:
            logs.info('create_subdomain_route', router_name, subdomain_route)
        else:
            routers_manager.create_subdomain_route(router_name, subdomain_route)
    if not dry_run:
        routers_manager.update(router_name, wait_ready=True)


def _get_namespace():
    return 'ckan-cloud'


def _get_frontend_hostname(storage_suffix=None):
    backend_url_target_id = _get_backend_url_target_id(storage_suffix=storage_suffix)
    routes = routers_manager.get_backend_url_routes(backend_url_target_id)
    assert storage_suffix or len(routes) == 1
    if len(routes) < 1:
        return 'localhost:9000'
    else:
        return routers_manager.get_route_frontend_hostname(routes[0])


def _get_backend_url_target_id(storage_suffix=None):
    return f'minio-{storage_suffix}' if storage_suffix else 'minio'



import click
import yaml

from ckan_cloud_operator import logs
from ckan_cloud_operator.config import manager


@click.group()
def config():
    pass


@config.command()
@click.option('--key')
@click.option('--default')
@click.option('--secret-name')
@click.option('--configmap-name')
@click.option('--namespace')
@click.option('--raw', is_flag=True)
@click.option('--template')
def get(**kwargs):
    """Get configuration values

    Examples:

        - get key foo from default configmap/namespace: get --key foo
        - get key foo from secret in default namespace: get --key foo --secret-name my-secret
        - get all keys from secret in given namespace: get --secret-name my-secret --namespace my-namespace
    """
    raw = kwargs.pop('raw', False)
    data = manager.get(**kwargs)
    if raw or kwargs.get('template'):
        print(data)
    else:
        print(yaml.dump(data, default_flow_style=False))


@config.command()
@click.argument('KEY')
@click.argument('VALUE')
@click.option('--secret-name')
@click.option('--configmap-name')
@click.option('--namespace')
@click.option('--from-file', is_flag=True)
def set(**kwargs):
    """Set a configuration value"""
    manager.set(**kwargs)
    logs.exit_great_success()


@config.command()
@click.argument('KEY')
@click.option('--secret-name')
@click.option('--namespace')
def delete_key(key, secret_name, namespace):
    """Delete a configuration value"""
    manager.delete_key(key, secret_name, namespace)
    logs.exit_great_success()


@config.command()
@click.option('--full', is_flag=True)
@click.option('--show-secrets', is_flag=True)
def list_configs(full, show_secrets):
    for config in manager.list_configs(full=full, show_secrets=show_secrets):
        print(yaml.dump([config], default_flow_style=False))



#### standard provider code ####

from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None, template=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix, template=template)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False, interactive=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file, interactive)


################################
# custom provider code starts here
#

import os
import binascii
import yaml
from ckan_cloud_operator import logs
from ckan_cloud_operator.drivers.rancher import driver as rancher_driver
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()


def initialize(interactive=False):
    _set_provider()
    assert interactive, 'non-interactive initialization is not supported'
    logs.info('To use the Rancher driver, please set the api endpoint credentials')
    print('You can get the credentials from Rancher UI, login as an admin user')
    print('  > Click on your profile pic top right corner'
          '    > API & Keys > Add Key')
    print('After you add the key the api endpoint and keys will be displayed')
    print('api-endpoint: https://<RANCHER_DOMAIN>/v3')
    print('default-context: <RANCHER_CLUSTER_ID>:<RANCHER_PROJECT_ID>')
    _config_interactive_set({
        'api-endpoint': '',
        'access-key': '',
        'secret-key': '',
        'default-context': '',

    }, is_secret=True, interactive=interactive)


def update(user):
    assert user['spec']['role'] == 'admin', 'only admin role is supported'
    username = user['spec']['name']
    password = _generate_password(12)
    config = _config_get(is_secret=True)
    api_endpoint, access_key, secret = [config['api-endpoint'], config['access-key'], config['secret-key']]
    print('---- # Creating Rancher user')
    res = rancher_driver.create_user(api_endpoint, access_key, secret, username, password)
    user_id = res['id']
    assert len(res['principalIds']) == 1, 'invalid principalIds'
    user_principal_id = res['principalIds'][0]
    print(yaml.dump(res, default_flow_style=False))
    cluster_id, _ = _config_get('default-context', is_secret=True).split(':')
    print('---- # Creating admin role bindings')
    logs.print_yaml_dump(
        rancher_driver.create_cluster_role_template_binding(api_endpoint, access_key, secret, cluster_id, user_principal_id)
    )
    print('---')
    logs.print_yaml_dump(
        rancher_driver.create_global_role_binding(api_endpoint, access_key, secret, user_id)
    )
    print('---- # Creating user token')
    res = rancher_driver.login(api_endpoint, username, password)
    user_login_token = res['token']
    logs.print_yaml_dump(res)
    print('---')
    res = rancher_driver.create_user_token(api_endpoint, user_login_token, 'ckan-cloud-operator')
    token = res['token']
    logs.print_yaml_dump(res)
    _config_set(f'{username}-token', token, is_secret=True)
    print('\n\nckan-cloud-operator uses the token to login, if you require human access, keep the username/password, it will not be displayed again\n\n')
    print('username:', username)
    print('password:', password)
    print('\n\n')
    logs.exit_great_success()


def get_kubeconfig(user):
    config = _config_get(is_secret=True)
    cluster_id, _ = config['default-context'].split(':')
    cluster_name = cluster_manager.get_cluster_name()
    api_endpoint = config['api-endpoint']
    k8s_endpoint = api_endpoint.replace('/v3', f'/k8s/clusters/{cluster_id}')
    user_name = user['spec']['name']
    user_token = config[f'{user_name}-token']
    return {
        "apiVersion": "v1",
        "kind": "Config",
        "clusters": [
            {
                "name": cluster_name,
                "cluster": {
                    "server": k8s_endpoint,
                    "api-version": "v1",

                }
            }
        ],
        "users": [
            {
                "name": user_name,
                "user": {
                    "token": user_token
                }
            }
        ],
        "contexts": [
            {
                "name": cluster_name,
                "context": {
                    "user": user_name,
                    "cluster": cluster_name
                }
            }
        ],
        "current-context": cluster_name
    }



from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.drivers.postgres import driver as postgres_driver

from ckan_cloud_operator.providers import manager as providers_manager
from ckan_cloud_operator.providers.db.proxy import manager as db_proxy_manager
from ckan_cloud_operator.providers.ckan import manager as ckan_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager

from .constants import PROVIDER_SUBMODULE as db_provider_submodule
from .azuresql.constants import PROVIDER_ID as db_azuresql_provider_id
from .gcloudsql.constants import PROVIDER_ID as db_gcloudsql_provider_id
from .rds.constants import PROVIDER_ID as db_rds_provider_id
from .minikube.constants import PROVIDER_ID as db_minikube_provider_id


def initialize(log_kwargs=None, interactive=False, default_cluster_provider=None):
    """Initialize / upgrade the db module and sub-modules"""
    if default_cluster_provider == 'aws':
        default_provider = db_rds_provider_id
    elif default_cluster_provider == 'azure':
        default_provider = db_azuresql_provider_id
    elif not default_cluster_provider or default_cluster_provider == 'gcloud':
        default_provider = db_gcloudsql_provider_id
    elif not default_cluster_provider or default_cluster_provider == 'minikube':
        default_provider = db_minikube_provider_id
    else:
        raise NotImplementedError(f'Unknown provider: {default_cluster_provider}')
    log_kwargs = log_kwargs or {}
    logs.info(f'Initializing DB provider', **log_kwargs)
    db_provider = providers_manager.get_provider(db_provider_submodule, default=default_provider)
    db_provider.initialize(interactive=interactive)
    if db_provider.is_private_ip():
        logs.info('DB Uses a private ip, initializing the DB proxy')
        db_proxy_manager.initialize()


def update():
    db_proxy_manager.update()


def get_all_db_names(db_prefix=None):
    return get_provider().get_all_db_namees(db_prefix=db_prefix)

def get_all_dbs_users():
    """Get a list of all databases and users, this is used by PgBouncer"""
    all_db_names, all_user_names, duplicate_db_names, duplicate_user_names = {}, {}, set(), set()
    all_dbs = []
    all_users = []
    for dbs_users in [
        _get_admin_dbs_users(),
        ckan_manager.get_all_dbs_users()
    ]:
        dbs, users = dbs_users
        for db in dbs:
            if all([s and s != 'None' for s in db]):
                db_name, _, _ = db
                if db_name in all_db_names:
                    if all_db_names[db_name] != db:
                        duplicate_db_names.add(db_name)
                else:
                    all_dbs.append(db)
                    all_db_names[db_name] = db
        for user in users:
            if all([s and s != 'None' for s in user]):
                user_name, _ = user
                if user_name in all_user_names:
                    if all_user_names[user_name] != user:
                        duplicate_user_names.add(user_name)
                else:
                    all_users.append(user)
                    all_user_names[user_name] = user
    assert len(duplicate_db_names) == 0 and len(duplicate_user_names) == 0, f'Found duplicate db/user names: {duplicate_db_names} / {duplicate_user_names}'
    return all_dbs, all_users


def get_admin_db_credentials(db_prefix=None):
    db_provider_manager = get_provider()
    admin_user, admin_password, db_name = db_provider_manager.get_postgres_admin_credentials(db_prefix=db_prefix)
    return admin_user, admin_password, db_name


def get_admin_db_user(db_prefix=None):
    admin_user, admin_password, db_name = get_admin_db_credentials(db_prefix=db_prefix)
    return admin_user


def get_external_admin_connection_string(db_name=None, db_prefix=None):
    """Get an admin connection string for access from outside the cluster"""
    admin_user, admin_password, admin_db_name = get_admin_db_credentials(db_prefix=db_prefix)
    if not db_name:
        db_name = admin_db_name
    db_host, db_port = get_external_proxy_host_port(db_prefix=db_prefix)
    return f'postgresql://{admin_user}:{admin_password}@{db_host}:{db_port}/{db_name}'


def get_deis_instsance_external_connection_string(instance_id, is_datastore=False, is_datastore_readonly=False, admin=False, db_prefix=None):
    user, password, db_name, db_host, db_port = get_deis_instance_external_connection_details(
        instance_id, is_datastore=is_datastore, is_datastore_readonly=is_datastore_readonly,
        db_prefix=db_prefix
    )
    if admin:
        user, password, _ = get_admin_db_credentials(db_prefix=db_prefix)
    return f'postgresql://{user}:{password}@{db_host}:{db_port}/{db_name}'


def get_external_connection_string(user, password, db_name, db_prefix=None):
    db_host, db_port = get_external_proxy_host_port(db_prefix=db_prefix)
    return f'postgresql://{user}:{password}@{db_host}:{db_port}/{db_name}'


def get_deis_instance_db_prefix(instance_id, is_datastore=False):
    instance_kind = ckan_manager.instance_kind()
    instance = kubectl.get(f'{instance_kind} {instance_id}', required=False)
    return get_deis_instance_db_prefix_from_instance(instance, is_datastore)


def get_deis_instance_db_prefix_from_instance(instance, is_datastore=False):
    return instance['spec'].get('datastore' if is_datastore else 'db', {}).get('dbPrefix') if instance else None


def get_default_db_prefix():
    from ckan_cloud_operator.config import manager as config_manager
    return config_manager.get('default-db-prefix', default='')


def get_deis_instance_credentials(instance_id, is_datastore=False, is_datastore_readonly=False, required=True,
                                  with_db_prefix=False):
    none = (None, None, None, None) if with_db_prefix else (None, None, None)
    instance_kind = ckan_manager.instance_kind()
    instance = kubectl.get(f'{instance_kind} {instance_id}', required=required)
    if not instance: return none
    secret = kubectl.get(f'secret {instance_id}-annotations', namespace=instance_id, required=required)
    if not secret: return none
    secret = kubectl.decode_secret(secret)
    if is_datastore or is_datastore_readonly:
        db_name = user = instance['spec'].get('datastore', {}).get('name')
        if is_datastore_readonly:
            user = secret.get('datastoreReadonlyUser')
            password = secret.get('datatastoreReadonlyPassword')
        else:
            password = secret.get('datastorePassword')
    else:
        db_name = user = instance['spec'].get('db', {}).get('name')
        password = secret.get('databasePassword')
    res = [user, password, db_name]
    if all(res):
        if with_db_prefix:
            res.append(get_deis_instance_db_prefix_from_instance(instance, is_datastore or is_datastore_readonly))
        return res
    else:
        assert not required, 'missing some db values'
        return none


def get_deis_instance_internal_connection_details(instance_id, is_datastore=False, is_datastore_readonly=False, required=True):
    user, password, db_name, db_prefix = get_deis_instance_credentials(instance_id, is_datastore, is_datastore_readonly, required,
                                                                       with_db_prefix=True)
    db_host, db_port = get_internal_proxy_host_port(db_prefix=db_prefix)
    res = [user, password, db_name, db_host, db_port]
    if all(res):
        return res
    else:
        assert not required, 'missing some db values'
        return None, None, None, None, None


def get_deis_instance_external_connection_details(instance_id, is_datastore=False, is_datastore_readonly=False, required=True, db_prefix=None):
    user, password, db_name, i_db_prefix = get_deis_instance_credentials(instance_id, is_datastore, is_datastore_readonly,
                                                                         required, with_db_prefix=True)
    if not db_prefix:
        db_prefix = i_db_prefix
    db_host, db_port = get_external_proxy_host_port(db_prefix=db_prefix)
    res = [user, password, db_name, db_host, db_port]
    if all(res):
        return res
    else:
        assert not required, 'missing some db values'
        return None, None, None, None, None


def get_internal_proxy_host_port(db_prefix=None):
    """Returns connection details for internal cluster access, via proxy if enabled"""
    db_proxy_provider = db_proxy_manager.get_provider(required=False)
    if db_proxy_provider:
        db_host, db_port = db_proxy_provider.get_internal_proxy_host_port(db_prefix=db_prefix)
    else:
        db_provider_manager = get_provider()
        db_host, db_port = db_provider_manager.get_postgres_internal_host_port(db_prefix=db_prefix)
    return db_host, db_port


def get_external_proxy_host_port(db_prefix=None):
    """Returns connection details for access from outside the cluster, via proxy if enabled"""
    from .gcloudsql.manager import _credentials_get

    is_private_ip = _credentials_get(db_prefix, key='is-private-ip', required=False) == 'y'
    if not is_private_ip:
        return get_provider().get_postgres_external_host_port(db_prefix=db_prefix)
    db_proxy_provider = db_proxy_manager.get_provider(required=False)
    assert db_proxy_provider, "SQL instance has private IP, so direct access to the DB is not supported, please enable the db proxy"
    host, port = db_proxy_provider.get_external_proxy_host_port(db_prefix=db_prefix)
    return (host, port) if host and port else get_provider().get_postgres_external_host_port(db_prefix=db_prefix)


def get_internal_unproxied_db_host_port(db_prefix=None):
    db_provider_manager = get_provider()
    return db_provider_manager.get_postgres_internal_host_port(db_prefix=db_prefix)


def get_provider():
    return providers_manager.get_provider(db_provider_submodule)


def check_db_exists(db_name, db_prefix=None):
    admin_connection_string = get_external_admin_connection_string(db_prefix=db_prefix)
    with postgres_driver.connect(admin_connection_string) as admin_conn:
        return len(list(postgres_driver.list_roles(admin_conn, role_name=db_name))) > 0


def check_connection_string(connection_string):
    with postgres_driver.connect(connection_string) as conn:
        with conn.cursor() as cur:
            cur.execute('select 1')


def _get_admin_dbs_users():
    db_provider_manager = get_provider()
    db_host, db_port = db_provider_manager.get_postgres_internal_host_port()
    db_admin_user, db_admin_password, db_admin_db = db_provider_manager.get_postgres_admin_credentials()
    dbs = [(db_admin_db, db_host, db_port)]
    users = [(db_admin_user, db_admin_password)]
    return dbs, users



import json
import requests
from urllib.parse import urljoin

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.drivers.helm import driver as helm_driver


def initialize(interactive=False):
    from .helm.manager import initialize as ckan_helm_initialize
    ckan_helm_initialize(interactive=interactive)


def update(instance_id, instance_type, instance, force=False, dry_run=False):
    return _get_deployment_provider(instance_type).update(instance_id, instance, force=force, dry_run=dry_run)


def delete(instance_id, instance_type, instance):
    if not instance_type:
        # for case of delete instance after instance object was already delete
        # we assume it's a helm deployment to do cleanup of the namepsace
        instance_type = 'helm'
    return _get_deployment_provider(instance_type).delete(instance_id, instance)


def get(instance_id, instance_type, instance):
    return _get_deployment_provider(instance_type).get(instance_id, instance)


def get_backend_url(instance_id, instance_type, instance):
    deployment_provider = _get_deployment_provider(instance_type, required=False)
    if deployment_provider:
        return deployment_provider.get_backend_url(instance_id, instance)
    else:
        return None


def pre_update_hook(instance_id, instance_type, instance, override_spec, skip_route, dry_run=False):
    return _get_deployment_provider(instance_type).pre_update_hook(instance_id, instance, override_spec, skip_route,
                                                                   dry_run=dry_run)


def get_deployment_version(instance_id):
    values = helm_driver.get_values(instance_id)
    values = json.loads(values)
    site_url = values.get('siteUrl')
    print('Current deployment version: ', requests.get(urljoin(site_url, 'version')).text)


def get_image(instance_id, service='ckan'):
    deployment_info = kubectl.get('deployment', namespace=instance_id)
    image_name = None
    for item in deployment_info.get('items', []):
        if item['metadata'].get('name') == service:
            containers = item.get('spec', {}).get('template', {}).get('spec', {}).get('containers', [])
            for container in containers:
                image_name = container.get('image')
                print(image_name)
    if image_name is None:
        print(f'Not able to find image for service "{service}", please make sure service name spelled correctly')


def set_image(instance_id, image_name, service='ckan', container_name=None):
    cont_name = container_name or service
    deployment_info = kubectl.call(
        f'set image deployment/{service} {cont_name}={image_name}',
        namespace=instance_id
    )


def create_ckan_admin_user(instance_id, instance_type, instance, user):
    _get_deployment_provider(instance_type).create_ckan_admin_user(instance_id, instance, user)


def _get_deployment_provider(instance_type, required=True):
    if instance_type == 'helm':
        from .helm import manager as helm_manager
        return helm_manager
    elif required:
        raise Exception(f'unknown instance_type ({instance_type})')
    else:
        return None



import click
import time

from ckan_cloud_operator import logs

from . import manager


@click.group()
def aws():
    """Manage an AWS cluster"""
    pass


@aws.command()
@click.argument('SUB_DOMAIN')
@click.argument('ROOT_DOMAIN')
@click.argument('LOAD_BALANCER_HOSTNAME')
def update_dns_record(sub_domain, root_domain, load_balancer_hostname):
    manager.update_dns_record(sub_domain, root_domain, load_balancer_hostname)
    logs.exit_great_success()



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import os
import yaml
import binascii

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from ckan_cloud_operator.drivers.gcloud import driver as gcloud_driver



def initialize(interactive=False):
    _set_provider()
    if interactive:
        print('\nEnter the path to a Google Cloud service account json file with relevant permissions\n')
        _config_interactive_set({'service-account-json': None}, is_secret=True, from_file=True)
        print('\nEnter the service account email\n')
        _config_interactive_set({'service-account-email': None}, is_secret=True)
        print('\nEnter the compute zone of the Google Kubernetes cluster which will be used as default compute zone\n')
        _config_interactive_set({'cluster-compute-zone': None})
        print('\nEnter the name of your cluster as defined in Google Kubernetes Engine\n')
        _config_interactive_set({'cluster-name': None})
        print('\nEnter the google project ID\n')
        _config_interactive_set({'project-id': None})

    gcloud_driver.activate_auth(
        _config_get('project-id'),
        _config_get('cluster-compute-zone'),
        _config_get('service-account-email', is_secret=True),
        _config_get('service-account-json', is_secret=True)
    )
    print(yaml.dump(get_info(), default_flow_style=False))


def activate_auth():
    gcloud_driver.activate_auth(
        _config_get('project-id'),
        _config_get('cluster-compute-zone'),
        _config_get('service-account-email', is_secret=True),
        _config_get('service-account-json', is_secret=True)
    )


def get_info(debug=False):
    cluster_name = _config_get('cluster-name')
    data = yaml.load(check_output(f'container clusters describe {cluster_name}'))
    if debug:
        return data
    else:
        return {
            'name': data['name'],
            'status': data['status'],
            'zone': data['zone'],
            'locations': data['locations'],
            'endpoint': data['endpoint'],
            'nodePools': [
                {
                    'name': pool['name'],
                    'status': pool['status'],
                    'version': pool['version'],
                    'config': {
                        'diskSizeGb': pool['config']['diskSizeGb'],
                        'machineType': pool['config']['machineType'],
                    },
                } for pool in data['nodePools']
            ],
            'createTime': data['createTime'],
            'currentMasterVersion': data['currentMasterVersion'],
            'currentNodeCount': data['currentNodeCount'],
        }


def get_name():
    return _config_get('cluster-name')


def get_cluster_kubeconfig_spec():
    cluster_name = _config_get('cluster-name')
    cluster = yaml.load(check_output(f'container clusters describe {cluster_name}'))
    return {
        "server": 'https://' + cluster['endpoint'],
        "certificate-authority-data": cluster['masterAuth']['clusterCaCertificate']
    }


def check_output(cmd, gsutil=False):
    return gcloud_driver.check_output(*get_project_zone(), cmd, gsutil=gsutil)


def check_call(cmd, gsutil=False):
    return gcloud_driver.check_call(*get_project_zone(), cmd, gsutil=gsutil)


def getstatusoutput(cmd, gsutil=False):
    return gcloud_driver.getstatusoutput(*get_project_zone(), cmd, gsutil=gsutil)


def get_project_zone():
    return _config_get('project-id'), _config_get('cluster-compute-zone')


def create_volume(disk_size_gb, labels, use_existing_disk_name=None, zone=None):
    disk_id = use_existing_disk_name or 'cc' + _generate_password(12)
    if use_existing_disk_name:
        logs.info(f'using existing persistent disk {disk_id}')
    else:
        logs.info(f'creating persistent disk {disk_id} with size {disk_size_gb}')
        _, zone = get_project_zone()
        labels = ','.join([
            '{}={}'.format(k.replace('/', '_'), v.replace('/', '_')) for k, v in labels.items()
        ])
        gcloud_driver.check_call(*get_project_zone(), f'compute disks create {disk_id} --size={disk_size_gb}GB --zone={zone} --labels={labels}')
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolume',
        'metadata': {'name': disk_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'storageClassName': '',
            'capacity': {'storage': f'{disk_size_gb}G'},
            'accessModes': ['ReadWriteOnce'],
            'gcePersistentDisk': {'pdName': disk_id}
        }
    })
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolumeClaim',
        'metadata': {'name': disk_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'storageClassName': '',
            'volumeName': disk_id,
            'accessModes': ['ReadWriteOnce'],
            'resources': {'requests': {'storage': f'{disk_size_gb}G'}}
        }
    })
    return {'persistentVolumeClaim': {'claimName': disk_id}}


def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False, suffix=None): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment, suffix=suffix)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import datetime
import os
import yaml
import subprocess
import json

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.cluster import manager as cluster_manager


def initialize(db_prefix=None, interactive=False):
    _set_provider()
    default_values = {
        'host': _get_resource_name(suffix=db_prefix) + '.ckan-cloud',
        'port': '5432',
        'is-private-ip': 'true',
        'admin-user': 'postgres',
        'admin-password': 'postgres',
    }
    config = _config_get(**_get_config_credentials_kwargs(db_prefix))
    for key, default_value in default_values.items():
        if key not in config and default_value:
            _config_set(key, default_value, **_get_config_credentials_kwargs(db_prefix))
        elif not config.get(key):
            raise Exception(f'missing key: {key}')
    _apply_service(db_prefix)
    _apply_deployment(db_prefix)


def get_postgres_internal_host_port(db_prefix=None):
    host = _credentials_get(db_prefix, key='host', required=True)
    port = int(_credentials_get(db_prefix, key='port', required=True))
    return host, port


def get_postgres_external_host_port(db_prefix=None):
    if os.environ.get('CKAN_CLOUD_OPERATOR_USE_PROXY') != 'n':
        assert _credentials_get(db_prefix, key='is-private-ip', required=False) != 'y', 'direct access to the DB is not supported, please enable the db proxy'
    host, port = get_postgres_internal_host_port(db_prefix)
    return host, port


def get_postgres_admin_credentials(db_prefix=None):
    credentials = _credentials_get(db_prefix)
    return credentials['admin-user'], credentials['admin-password'], credentials.get('admin-db-name', credentials['admin-user'])


def is_private_ip(db_prefix=None):
    return _credentials_get(db_prefix, key='is-private-ip', required=False) == 'y'


def _credentials_get(db_prefix, key=None, default=None, required=False):
    return _config_get(key=key, default=default, required=required, **_get_config_credentials_kwargs(db_prefix))


def _get_config_credentials_kwargs(db_prefix):
    return {
        'is_secret': True,
        'suffix': f'{db_prefix}-credentials' if db_prefix else 'credentials'
    }



def _apply_deployment(db_prefix=None):
    kubectl.apply(kubectl.get_deployment(
        _get_resource_name(suffix=db_prefix),
        _get_resource_labels(for_deployment=True, suffix=db_prefix or ''),
        {
            'replicas': 1,
            'revisionHistoryLimit': 10,
            'strategy': {'type': 'RollingUpdate', },
            'selector': {
                'matchLabels': _get_resource_labels(for_deployment=True, suffix=db_prefix or '')
            },
            'template': {
                'metadata': {
                    'labels': _get_resource_labels(for_deployment=True, suffix=db_prefix or ''),
                    'annotations': _get_resource_annotations(suffix=db_prefix or '')
                },
                'spec': {
                    'containers': [
                        {
                            'name': 'postgres',
                            'image': 'mdillon/postgis',
                            'env': [
                                {
                                    'name': 'POSTGRES_PASSWORD',
                                    'value': 'postgres'
                                },
                                {
                                    'name': 'POSTGRES_USER',
                                    'value': 'postgres'
                                },
                            ],
                            'ports': [{'containerPort': 5432}],
                            'resources': {
                                'limits': {
                                    'memory': '1Gi',
                                },
                                'requests': {
                                    'cpu': '0.1',
                                    'memory': '0.2Gi',
                                }
                            }
                        }
                    ],
                }
            }
        }
    ))


def _apply_service(db_prefix=None):
    deployment_app = _get_resource_labels(for_deployment=True, suffix=db_prefix or '')['app']
    kubectl.apply(kubectl.get_service(
        _get_resource_name(suffix=db_prefix or ''),
        _get_resource_labels(suffix=db_prefix or ''),
        [5432],
        {'app': deployment_app}
    ))



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import datetime
import os
import yaml
import subprocess
import json

from ckan_cloud_operator import logs
from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.drivers.gcloud import driver as gcloud_driver


def initialize(db_prefix=None, interactive=False):
    _set_provider()
    default_values = {
        'rds-instance-name': '',
        'rds-host': '',
        'admin-user': '',
        'admin-password': '',
    }
    if interactive:
        print("\n"
              "Starting interactive initialization of the AWS RDS db provider\n"
              "Please prepare the following values:\n"
              "\n"
              " - admin db host and credentials\n"
              " - RDS instance name\n"
              "\n")
        _config_interactive_set(default_values, **_get_config_credentials_kwargs(db_prefix))
    config = _config_get(**_get_config_credentials_kwargs(db_prefix))
    for key, default_value in default_values.items():
        if key not in config and default_value:
            _config_set(key, default_value, **_get_config_credentials_kwargs(db_prefix))
        elif not config.get(key):
            raise Exception(f'missing key: {key}')
    from ckan_cloud_operator.providers.db.proxy.rds import manager as db_proxy_rds_manager
    db_proxy_rds_manager.initialize(db_prefix=db_prefix)


def get_postgres_internal_host_port(db_prefix=None):
    host = _credentials_get(db_prefix, key='rds-host', required=True)
    port = 5432
    return host, port


def get_postgres_external_host_port(db_prefix=None):
    if os.environ.get('CKAN_CLOUD_OPERATOR_USE_PROXY') != 'n':
        assert _credentials_get(db_prefix, key='is-private-ip', required=False) != 'y', 'direct access to the DB is not supported, please enable the db proxy'
    host, port = get_postgres_internal_host_port(db_prefix)
    return host, port


def get_postgres_admin_credentials(db_prefix=None):
    credentials = _credentials_get(db_prefix)
    return credentials['admin-user'], credentials['admin-password'], credentials.get('admin-db-name', credentials['admin-user'])


def is_private_ip(db_prefix=None):
    return _credentials_get(db_prefix, key='is-private-ip', required=False) == 'y'


def import_db(import_url, target_db_name, import_user, db_prefix=None):
    gcloud_driver._import_gcloud_sql_db(
        *_gcloud().get_project_zone(),
        _sql_instance_name(db_prefix),
        import_url, target_db_name, import_user
    )


def get_all_db_names(db_prefix=None):
    instance_name = _sql_instance_name(db_prefix)
    return [
        db['name'] for db
        in json.loads(gcloud_driver.check_output(
            *_gcloud().get_project_zone(),
            f'sql databases list --instance {instance_name} --format json'
        ).decode())
    ]


def create_backup(database, connection_string=None, db_prefix=None, if_not_exists=False, dry_run=False):
    filename = f'{database}_' + datetime.datetime.now().strftime('%Y%m%d%H%M') + '.gz'
    gs_url = os.path.join(
        _credentials_get(None, key='backups-gs-base-url', required=True),
        *([db_prefix] if db_prefix else []),
        datetime.datetime.now().strftime('%Y/%m/%d/%H'),
        filename
    )
    if if_not_exists and gcloud_driver.call(*_gcloud().get_project_zone(), f'ls {gs_url}') == 0:
        return False
    else:
        if not connection_string:
            from ckan_cloud_operator.providers.db import manager as db_manager
            connection_string = db_manager.get_external_admin_connection_string(db_name=database, db_prefix=db_prefix)
        logs.info(f'Dumping DB: {filename}')
        if not dry_run:
            subprocess.check_call([
                "bash", "-o", "pipefail", "-c",
                    f"pg_dump -d {connection_string} --format=plain --no-owner --no-acl --schema=public | "
                    f"sed -E 's/(DROP|CREATE|COMMENT ON) EXTENSION/-- \\1 EXTENSION/g' | "
                    f"gzip -c > {filename}",
            ])
            subprocess.check_call(f'ls -lah {filename}', shell=True)
        logs.info(f'Copying to: {gs_url}')
        if not dry_run:
            gcloud_driver.check_call(
                *_gcloud().get_project_zone(),
                f'-m cp ./{filename} {gs_url} && rm {filename}',
                gsutil=True
            )
        return True


def create_all_backups(db_prefix=None, dry_run=False):
    logs.info('Fetching all database names')
    db_names = get_all_db_names(db_prefix=db_prefix)
    logs.info('{} DBs'.format(len(db_names)))
    for db in db_names:
        if db not in ['postgres'] and not db.startswith('template'):
            assert create_backup(db, db_prefix=db_prefix, dry_run=dry_run)


def get_operation_status(operation_id):
    return yaml.load(gcloud_driver.check_output(
        *_gcloud().get_project_zone(),
        f'sql operations describe {operation_id}'
    ).decode())


def _credentials_get(db_prefix, key=None, default=None, required=False):
    return _config_get(key=key, default=default, required=required, **_get_config_credentials_kwargs(db_prefix))


def _get_config_credentials_kwargs(db_prefix):
    return {
        'is_secret': True,
        'suffix': f'{db_prefix}-credentials' if db_prefix else 'credentials'
    }


def _gcloud():
    return cluster_manager.get_provider()


def _sql_instance_name(db_prefix):
    return _credentials_get(db_prefix, key='gcloud-sql-instance-name')



import subprocess
import yaml

from ckan_cloud_operator.routers import manager as routers_manager
from ckan_cloud_operator.routers.routes import manager as routes_manager
from ckan_cloud_operator import logs


def add_cli_commands(click, command_group, great_success):

    @command_group.command('initialize')
    @click.option('--interactive', is_flag=True)
    def initialize(interactive):
        from ckan_cloud_operator.providers.routers.manager import initialize as manager_initialize
        manager_initialize(interactive)
        great_success()

    @command_group.command('create-traefik-router')
    @click.argument('TRAEFIK_ROUTER_NAME')
    @click.argument('DEFAULT_ROOT_DOMAIN', required=False, default='default')
    @click.argument('CLOUDFLARE_EMAIL', required=False, default='default')
    @click.argument('CLOUDFLARE_API_KEY', required=False, default='default')
    @click.option('--external-domains', is_flag=True)
    def routers_create(traefik_router_name, default_root_domain, cloudflare_email, cloudflare_api_key, external_domains):
        """Create a Traefik router, with domain registration and let's encrypt based on Cloudflare"""
        routers_manager.create(
            traefik_router_name,
            routers_manager.get_traefik_router_spec(
                default_root_domain, cloudflare_email, cloudflare_api_key,
                external_domains=external_domains
            )
        )
        routers_manager.update(traefik_router_name)
        great_success()

    @command_group.command('update')
    @click.argument('ROUTER_NAME')
    @click.option('--wait-ready', is_flag=True)
    def routers_update(router_name, wait_ready):
        """Update a router to latest resource spec"""
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('list')
    @click.option('-f', '--full', is_flag=True)
    @click.option('-v', '--values-only', is_flag=True)
    def routers_list(**kwargs):
        """List the router resources"""
        routers_manager.list(**kwargs)

    @command_group.command('kubectl-get-all')
    @click.argument('ROUTER_TYPE', default=routers_manager.DEFAULT_ROUTER_TYPE)
    def routers_kubectl_get_all(router_type):
        subprocess.check_call(f'kubectl -n ckan-cloud get all -l ckan-cloud/router-type={router_type}',
                              shell=True)

    @command_group.command('create-deis-instance-subdomain-route')
    @click.argument('ROUTER_NAME')
    @click.argument('DEIS_INSTANCE_ID')
    @click.argument('SUB_DOMAIN', required=False, default='default')
    @click.argument('ROOT_DOMAIN', required=False, default='default')
    @click.option('--wait-ready', is_flag=True)
    def routers_create_deis_instance_subdomain_route(router_name, deis_instance_id,
                                                     sub_domain, root_domain,
                                                     wait_ready):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'deis-instance',
            'deis-instance-id': deis_instance_id,
            'root-domain': root_domain,
            'sub-domain': sub_domain
        })
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('create-ckan-instance-subdomain-route')
    @click.argument('ROUTER_NAME')
    @click.argument('CKAN_INSTANCE_ID')
    @click.argument('SUB_DOMAIN', required=False, default='default')
    @click.argument('ROOT_DOMAIN', required=False, default='default')
    @click.option('--wait-ready', is_flag=True)
    def routers_create_ckan_instance_subdomain_route(router_name, ckan_instance_id,
                                                     sub_domain, root_domain,
                                                     wait_ready):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'ckan-instance',
            'ckan-instance-id': ckan_instance_id,
            'root-domain': root_domain,
            'sub-domain': sub_domain
        })
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('create-app-instance-subdomain-route')
    @click.argument('ROUTER_NAME')
    @click.argument('APP_INSTANCE_ID')
    @click.argument('SUB_DOMAIN', required=False, default='default')
    @click.argument('ROOT_DOMAIN', required=False, default='default')
    @click.option('--wait-ready', is_flag=True)
    def routers_create_app_instance_subdomain_route(router_name, app_instance_id,
                                                    sub_domain, root_domain,
                                                    wait_ready):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'app-instance',
            'app-instance-id': app_instance_id,
            'root-domain': root_domain,
            'sub-domain': sub_domain
        })
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('create-datapusher-subdomain-route')
    @click.argument('ROUTER_NAME')
    @click.argument('DATAPUSHER_NAME')
    @click.argument('SUB_DOMAIN', required=False, default='default')
    @click.argument('ROOT_DOMAIN', required=False, default='default')
    @click.option('--wait-ready', is_flag=True)
    def routers_create_datapusher_subdomain_route(router_name, datapusher_name,
                                                  sub_domain, root_domain,
                                                  wait_ready):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'datapusher',
            'datapusher-name': datapusher_name,
            'root-domain': root_domain,
            'sub-domain': sub_domain
        })
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('create-backend-url-subdomain-route')
    @click.argument('ROUTER_NAME')
    @click.argument('TARGET_RESOURCE_ID')
    @click.argument('BACKEND_URL')
    @click.argument('SUB_DOMAIN', required=False, default='default')
    @click.argument('ROOT_DOMAIN', required=False, default='default')
    @click.option('--wait-ready', is_flag=True)
    @click.option('--httpauth-secret')
    def routers_create_backend_url_subdomain_route(router_name, target_resource_id, backend_url,
                                                   sub_domain, root_domain, wait_ready,
                                                   httpauth_secret):
        routers_manager.create_subdomain_route(router_name, {
            'target-type': 'backend-url',
            'target-resource-id': target_resource_id,
            'backend-url': backend_url,
            'sub-domain': sub_domain,
            'root-domain': root_domain,
            **({'httpauth-secret': httpauth_secret} if httpauth_secret else {}),
        })
        routers_manager.update(router_name, wait_ready)
        great_success()

    @command_group.command('get-routes')
    @click.option('-p', '--datapusher-name', required=False)
    @click.option('-d', '--deis-instance-id', required=False)
    @click.option('--ckan-instance-id', required=False)
    @click.option('-b', '--backend-url-target-id', required=False)
    @click.option('-o', '--one', is_flag=True)
    @click.option('-e', '--external-domain', is_flag=True)
    @click.option('--edit', is_flag=True)
    def get_routes(datapusher_name, deis_instance_id, ckan_instance_id, backend_url_target_id, one, external_domain, edit):
        if datapusher_name:
            assert not deis_instance_id and not ckan_instance_id and not backend_url_target_id
            routes = routers_manager.get_datapusher_routes(datapusher_name, edit=edit)
        elif deis_instance_id:
            assert not datapusher_name and not ckan_instance_id and not backend_url_target_id
            routes = routers_manager.get_deis_instance_routes(deis_instance_id, edit=edit)
        elif ckan_instance_id:
            assert not datapusher_name and not deis_instance_id and not backend_url_target_id
            routes = routers_manager.get_ckan_instance_routes(ckan_instance_id, edit=edit)
        elif backend_url_target_id:
            assert not datapusher_name and not deis_instance_id and not ckan_instance_id
            routes = routers_manager.get_backend_url_routes(backend_url_target_id, edit=edit)
        else:
            routes = routers_manager.get_all_routes()
        if routes:
            if one: assert len(routes) == 1, 'too many routes!'
            for route in routes:
                if external_domain:
                    data = routers_manager.get_route_frontend_hostname(route)
                    if one:
                        print(data)
                    else:
                        print(yaml.dump([data], default_flow_style=False))
                else:
                    try:
                        data = {
                            'name': route['metadata']['name'],
                            'backend-url': routes_manager.get_backend_url(route),
                            'frontend-hostname': routes_manager.get_frontend_hostname(route),
                            'router-name': route['spec']['router_name']
                        }
                    except Exception as e:
                        print('Warning: %s' % repr(e))
                        continue
                    if one:
                        print(yaml.dump(data, default_flow_style=False))
                    else:
                        print(yaml.dump([data], default_flow_style=False))

    @command_group.command('delete-routes')
    @click.option('-p', '--datapusher-name', required=False)
    @click.option('-d', '--deis-instance-id', required=False)
    @click.option('-r', '--root-domain', required=False)
    @click.option('-s', '--sub-domain', required=False)
    def delete_routes(datapusher_name, deis_instance_id, root_domain, sub_domain):
        routers_manager.delete_routes(
            datapusher_name=datapusher_name,
            deis_instance_id=deis_instance_id,
            root_domain=root_domain,
            sub_domain=sub_domain
        )
        logs.exit_great_success()

    @command_group.command('delete')
    @click.argument('ROUTER_NAME')
    @click.argument('ROUTER_TYPE', required=False, default='')
    def routers_delete(router_name, router_type):
        routers_manager.delete(router_name, router_type)
        great_success()

    @command_group.command('get')
    @click.argument('ROUTER_NAME')
    @click.option('--dns', is_flag=True)
    def get(router_name, dns):
        print(yaml.dump(routers_manager.get(router_name, only_dns=dns), default_flow_style=False))

    @command_group.command('cloudflare-rate-limits')
    @click.argument('ROOT_DOMAIN')
    def cloudflare_rate_limits(root_domain):
        logs.print_yaml_dump(routers_manager.get_cloudflare_rate_limits(root_domain))



import click
import yaml
import json

from ckan_cloud_operator import logs

from ckan_cloud_operator.providers.ckan.db import migration as db_migration_manager
from ckan_cloud_operator.providers.ckan import manager
from .storage import cli as ckan_storage_cli
from .deployment import cli as ckan_deployment_cli
from .instance import cli as instance_cli
from .env import cli as env_cli

@click.group()
def ckan():
    """Manage CKAN Instances"""
    pass


# ckan.add_command(ckan_storage_cli.storage)
ckan.add_command(ckan_deployment_cli.deployment)
ckan.add_command(instance_cli.instance)
ckan.add_command(env_cli.env)

@ckan.command()
@click.option('--environment-name', help='name of the environment to initialize (one of the result of `cco ckan env list`)')
def init(environment_name):
    '''
    Initialize ckan-cloud-operator for working with environment.
    This command gets all the necessary credentials for working
    with the given environment. Eg, gets and saves `kubeconfig`
    file in `~/.kube/config` directory

    Without flags initializes current working environment added with `cco ckan env add`

    \b
    cco ckan init --environment-name poc
    > POC environment was succesfully initialized
    '''

############################################################################
## Commands Below are not used for now and are making unnecessary noise   ##
## I'm commenting them for now, but not deleting as they might be useful  ##
## In future. Will bring them back as needed                              ##
############################################################################
#
# @ckan.command()
# @click.option('--interactive', is_flag=True)
# def initialize(interactive):
#     manager.initialize(interactive=interactive)
#     logs.exit_great_success()
#
#
# @ckan.command()
# @click.argument('OLD_SITE_ID')
# @click.argument('NEW_INSTANCE_ID', required=False)
# @click.argument('ROUTER_NAME', required=False)
# @click.option('--skip-gitlab', is_flag=True)
# @click.option('--force', is_flag=True)
# @click.option('--rerun', is_flag=True)
# @click.option('--recreate-dbs', is_flag=True)
# @click.option('--recreate', is_flag=True)
# @click.option('--recreate-instance', is_flag=True)
# @click.option('--skip-routes', is_flag=True)
# @click.option('--skip-solr', is_flag=True)
# @click.option('--skip-deployment', is_flag=True)
# @click.option('--no-db-proxy', is_flag=True)
# def migrate_deis_instance(old_site_id, new_instance_id, router_name, skip_gitlab, force, rerun, recreate_dbs, recreate,
#                           recreate_instance, skip_routes, skip_solr, skip_deployment, no_db_proxy):
#     """Run a full end-to-end migration of an instasnce"""
#     manager.migrate_deis_instance(old_site_id, new_instance_id, router_name, skip_gitlab, force, rerun, recreate_dbs,
#                                   recreate, recreate_instance, skip_routes, skip_solr, skip_deployment, no_db_proxy)
#     logs.exit_great_success()
#
#
# @ckan.command()
# @click.argument('OLD_SITE_ID')
# @click.option('--db-name')
# @click.option('--datastore-name')
# @click.option('--force', is_flag=True)
# @click.option('--rerun', is_flag=True)
# @click.option('--recreate-dbs', is_flag=True)
# @click.option('--dbs-suffix')
# @click.option('--skip-create-dbs', is_flag=True)
# @click.option('--skip-datastore-import', is_flag=True)
# def migrate_deis_dbs(old_site_id, db_name, datastore_name, force, rerun, recreate_dbs, dbs_suffix, skip_create_dbs,
#                      skip_datastore_import):
#     migration_generator = db_migration_manager.migrate_deis_dbs(
#         old_site_id, db_name, datastore_name, force=force, rerun=rerun, recreate_dbs=recreate_dbs, dbs_suffix=dbs_suffix,
#         skip_create_dbs=skip_create_dbs, skip_datastore_import=skip_datastore_import
#     )
#     for event in migration_generator:
#         db_migration_manager.print_event_exit_on_complete(
#             event,
#             f'{old_site_id} -> {db_name}, {datastore_name}'
#         )
#
#
# @ckan.command()
# @click.option('--db-name')
# def migrate_list(db_name):
#     for item in db_migration_manager.get()['items']:
#         data = dict(item.get('spec', {}), **{'resource-name': item.get('metadata', {}).get('name')})
#         if data.get('db-name') == db_name or not db_name:
#             print(yaml.dump([data], default_flow_style=False))
#
#
# @ckan.command()
# @click.argument('MIGRATION_NAME', nargs=-1)
# @click.option('--delete-dbs', is_flag=True)
# def migrate_delete(migration_name, delete_dbs):
#     for name in migration_name:
#         db_migration_manager.delete(name, delete_dbs)
#     logs.exit_great_success()
#
#
# @ckan.command()
# def get_all_dbs_users():
#     dbs, users = manager.get_all_dbs_users()
#     print(yaml.dump({
#         'dbs': [' | '.join(map(str, db)) for db in dbs],
#         'users': [' | '.join(map(str, user)) for user in users]
#     }, default_flow_style=False))
#
#
# @ckan.command()
# @click.argument('INSTANCE_ID')
# def post_create_checks(instance_id):
#     manager.post_create_checks(instance_id)
#     logs.exit_great_success()
#
#
# @ckan.command()
# @click.argument('INSTANCE_ID')
# def admin_credentials(instance_id):
#     logs.print_yaml_dump(manager.ckan_admin_credentials(instance_id))
#
#
# @ckan.command()
# @click.argument('OLD_SITE_ID')
# @click.option('-r', '--raw', is_flag=True)
# def db_migration_import_urls(old_site_id, raw):
#     urls = db_migration_manager.get_db_import_urls(old_site_id)
#     if raw:
#         print(' '.join(urls))
#     else:
#         logs.print_yaml_dump(list(urls))
